/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/crypt/bfish.c                                            *
 * Created:     2000-08-24 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2000-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>
#include <lib/md5.h>

#ifdef MISC_DOS
#include <time.h>
#include <bios.h>
#else
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#endif

#include "bfish.h"
#include "blowfish.h"
#include "ofb64.h"
#include "passwd.h"


#define BFISH_MAGIC 0x42465348UL

#define BFISH_BUF_SIZE MISC_BUF_SIZE


/*
File format:
0       4   Magic ("BFSH")
4       4   Format version (0)
8       8   IV
16      n   encrypted data
n + 16  16  encrypted MD5 sum of IV + unencrypted data
*/

static const char    *argv0 = NULL;
static int           par_enc = 1;
static int           par_verbose = 0;
static unsigned long par_iv = 0;
static int           par_stdin = 0;
static char          *par_prompt = NULL;
static unsigned      par_pwcnt = 0;


static
void prt_help (void)
{
	fputs (
		"Usage: bfish [options] [input [output]]\n"
		"  -i string                 Set input file name [stdin]\n"
		"  -o string                 Set output file name [stdout]\n"
		"  -e, --encrypt             Encrypt [default]\n"
		"  -d, --decrypt             Decrypt\n"
		"  -j, --iv int              Set unique part of iv (12 bits) [0]\n"
		"  -n, --password-count int  Read the password n times [0]\n"
		"  -p, --prompt string       Set the password prompt string\n"
		"  -s, --stdin               Read the password from stdin\n",
		stdout
	);

	fflush (stdout);
}

static
void bfish_set_uint32 (void *buf, unsigned i, unsigned long v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = (v >> 24) & 0xff;
	tmp[1] = (v >> 16) & 0xff;
	tmp[2] = (v >> 8) & 0xff;
	tmp[3] = v & 0xff;
}

static
unsigned long bfish_get_uint32 (const void *buf, unsigned i)
{
	unsigned long       v;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	v = tmp[0];
	v = (v << 8) | tmp[1];
	v = (v << 8) | tmp[2];
	v = (v << 8) | tmp[3];

	return (v);
}

#ifdef MISC_DOS

static
int bfish_get_timestamp (unsigned long *sec, unsigned long *usec)
{
	*sec = (unsigned long) time (NULL);
	if (*sec < BFISH_20000101) {
		*sec = BFISH_20000101;
	}
	else {
		*sec -= BFISH_20000101;
	}

	*usec = biostime (0, 0);

	return (0);
}

#else

static
int bfish_get_timestamp (unsigned long *sec, unsigned long *usec)
{
	struct timeval tv;

	if (gettimeofday (&tv, NULL) < 0) {
		return (1);
	}

	*sec = (unsigned long) tv.tv_sec;
	if (*sec < BFISH_20000101) {
		*sec = BFISH_20000101;
	}
	else {
		*sec -= BFISH_20000101;
	}

	*usec = (unsigned long) tv.tv_usec;

	return (0);
}

#endif

static
int bfish_get_iv (block64_t *iv)
{
	unsigned long  hi, lo;

	if (bfish_get_timestamp (&lo, &hi)) {
		return (1);
	}

	hi = ((hi & 0xfffffUL) | (par_iv << 20)) & 0xffffffffUL;

	if (par_verbose) {
		fprintf (stderr, "IV1: %08lx%08lx\n", hi, lo);
	}

	blk64_set_32 (iv, 0, hi);
	blk64_set_32 (iv, 1, lo);

	return (0);
}

static
int bfish_get_key (bfish_ctx_t *bfish, unsigned cnt)
{
	unsigned      i, n, m;
	unsigned char buf[256];
	unsigned char tmp[256];

	if (par_prompt == NULL) {
		par_prompt = "bfish password: ";
	}

	if (par_stdin) {
		n = 256;

		if (crypt_get_passwd_fp (stdin, buf, &n)) {
			return (1);
		}
	}
	else {
		i = 0;
		while (i < cnt) {
			if (i == 0) {
				n = 256;

				if (crypt_get_passwd (par_prompt, buf, &n)) {
					return (1);
				}

				i += 1;
			}
			else {
				m = 256;

				if (crypt_get_passwd (par_prompt, tmp, &m)) {
					return (1);
				}

				if ((n != m) || (memcmp (buf, tmp, n) != 0)) {
					fputs ("password mismatch\n", stderr);
					fflush (stderr);
					i = 0;
				}
				else {
					i += 1;
				}
			}
		}
	}

	bfish_set_key (bfish, buf, n);

	return (0);
}

static
int bfish_ofb_enc (FILE *inp, FILE *out)
{
	unsigned long n;
	unsigned char buf[BFISH_BUF_SIZE + 16];
	block64_t     iv;
	md5_t         md5;
	bfish_ctx_t   bfish;
	ofb_ctx_t     ofb;

	if (bfish_get_key (&bfish, (par_pwcnt > 0) ? par_pwcnt : 2)) {
		return (1);
	}

	md5_init (&md5);
	ofb_init (&ofb, &bfish_crypt, &bfish);

	bfish_set_uint32 (buf, 0, BFISH_MAGIC);
	bfish_set_uint32 (buf, 4, 0);

	bfish_get_iv (&iv);
	bfish_crypt (&bfish, &iv, 1);
	blk64_set_buf (&iv, &buf[8]);

	ofb_set_iv_blk (&ofb, &iv);

	md5_calc (&md5, &buf[8], 8);

	if (fwrite (buf, 1, 16, out) != 16) {
		return (1);
	}

	if (par_verbose) {
		fprintf (stderr, "IV2: %08lx%08lx\n", iv.word32[0], iv.word32[1]);
	}

	n = fread (buf, 1, BFISH_BUF_SIZE, inp);

	while (n > 0) {
		md5_calc (&md5, buf, n);
		ofb_crypt (&ofb, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = fread (buf, 1, BFISH_BUF_SIZE, inp);
	}

	md5_done (&md5);
	md5_get_sum (&md5, buf);
	ofb_crypt (&ofb, buf, 16);

	if (fwrite (buf, 1, 16, out) != 16) {
		return (1);
	}

	fflush (out);

	return (0);
}

static
int bfish_ofb_dec (FILE *inp, FILE *out)
{
	unsigned long i, n;
	unsigned long vers;
	unsigned char buf[BFISH_BUF_SIZE + 16];
	md5_t         md5;
	bfish_ctx_t   bfish;
	ofb_ctx_t     ofb;

	if (fread (buf, 1, 16, inp) != 16) {
		return (1);
	}

	if (bfish_get_uint32 (buf, 0) != BFISH_MAGIC) {
		fprintf (stderr, "%s: bad magic\n", argv0);
		return (1);
	}

	vers = bfish_get_uint32 (buf, 4);
	if (vers != 0) {
		fprintf (stderr, "%s: bad version number (%lu)\n", argv0, vers);
		return (1);
	}

	if (bfish_get_key (&bfish, (par_pwcnt > 0) ? par_pwcnt : 1)) {
		return (1);
	}

	md5_init (&md5);
	md5_calc (&md5, &buf[8], 8);

	ofb_init (&ofb, &bfish_crypt, &bfish);
	ofb_set_iv_buf (&ofb, &buf[8]);

	n = fread (buf, 1, BFISH_BUF_SIZE + 16, inp);
	if (n < 16) {
		fprintf (stderr, "%s: input file too short\n", argv0);
		return (1);
	}

	ofb_crypt (&ofb, buf, n);

	n -= 16;

	while (n > 0) {
		md5_calc (&md5, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		for (i = 0; i < 16; i++) {
			buf[i] = buf[n + i];
		}

		n = fread (buf + 16, 1, BFISH_BUF_SIZE, inp);
		ofb_crypt (&ofb, buf + 16, n);
	}

	fflush (out);

	md5_done (&md5);
	md5_get_sum (&md5, buf + 16);

	for (i = 0; i < 16; i++) {
		if (buf[i] != buf[i + 16]) {
			fprintf (stderr, "%s: MD5 mismatch\n", argv0);
			return (1);
		}
	}

	return (0);
}

static
int do_bfish (FILE *inp, FILE *out)
{
	int r;

	if (par_enc) {
		r = bfish_ofb_enc (inp, out);
	}
	else {
		r = bfish_ofb_dec (inp, out);
	}

	return (r);
}

int main (int argc, char *argv[])
{
	int        i, r;
	const char *finp;
	const char *fout;
	FILE       *inp;
	FILE       *out;
	unsigned   fcnt;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help ();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("bfish");
			return (EXIT_SUCCESS);
		}
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "e", "encrypt")) {
			par_enc = 1;
		}
		else if (str_isarg2 (argv[i], "d", "deccrypt")) {
			par_enc = 0;
		}
		else if (str_isarg2 (argv[i], "p", "prompt")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing prompt\n", argv0);
				return (EXIT_FAILURE);
			}

			par_prompt = argv[i];
		}
		else if (str_isarg2 (argv[i], "v", "verbose")) {
			par_verbose = 1;
		}
		else if (str_isarg2 (argv[i], "j", "iv")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing iv\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_iv)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "n", "password-count")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing password count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_pwcnt)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "s", "stdin")) {
			par_stdin = 1;
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown argument (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if ((finp == NULL) && par_stdin) {
		fprintf (stderr, "%s: can't read password from stdin\n", argv0);
		return (EXIT_FAILURE);
	}

	if (finp == NULL) {
		inp = stdin;
	}
	else {
		inp = fopen (finp, "rb");
		if (inp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
			return (EXIT_FAILURE);
		}
	}

	if (fout == NULL) {
		out = stdout;
	}
	else {
		out = fopen (fout, "wb");
		if (out == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
			return (EXIT_FAILURE);
		}
	}

	r = do_bfish (inp, out);

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
