/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/fcmp/fcmp.c                                              *
 * Created:     2000-05-12 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2000-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/file.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#if defined(HAVE_TIME_H) && defined(HAVE_UNISTD_H) && \
	defined(HAVE_SYS_TYPES_H) && defined(HAVE_SYS_STAT_H)
#define FCMP_STAT 1
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "fcmp.h"


static const char    *argv0 = NULL;

static int           par_quiet = 0;
static int           par_header = 0;
static unsigned long par_diff_max = 0;

static unsigned long par_file_cnt = 0;
static fcmp_t        *par_files = NULL;


static
void prt_help (void)
{
	fputs (
		"fcmp: compare binary files\n\n"
		"usage: fcmp [options] [filename...]\n"
		"  -f, --first       Compare only up to the first difference [no]\n"
		"  -h, --header      Print a header [no]\n"
		"  -n, --max-diff n  Stop after n different bytes [0]\n"
		"  -q, --quiet       No output to stdout [no]\n",
		stdout
	);

	fflush (stdout);
}

static
int fcmp_new_file (const char *fname)
{
	file_t *f;

	par_files = (fcmp_t *) realloc (par_files, sizeof (fcmp_t) * (par_file_cnt + 1));
	if (par_files == NULL) {
		return (1);
	}

	file_init (&par_files[par_file_cnt].f);
	par_files[par_file_cnt].chr = 0;
	par_files[par_file_cnt].eof = 0;
	par_files[par_file_cnt].fname = NULL;

	f = &par_files[par_file_cnt].f;

	if (strcmp (fname, "-") == 0) {
		file_set_stdin (f);
	}
	else {
		file_set_name (f, fname);
		par_files[par_file_cnt].fname = strdup (fname);
	}

	if (file_open_inp (f)) {
		return (1);
	}

	par_file_cnt += 1;

	return (0);
}

static
void fcmp_close (void)
{
	unsigned long i;

	for (i = 0; i < par_file_cnt; i++) {
		file_close (&par_files[i].f);
		free (par_files[i].fname);
	}
}

/* read next characters. returns 1 if all files have reached eof. */
static
int fcmp_next_char (void)
{
	unsigned long i;
	int           eof;

	eof = 1;

	for (i = 0; i < par_file_cnt; i++) {
		if (par_files[i].eof == 0) {
			if (file_get_byte (&par_files[i].f, &par_files[i].chr)) {
				par_files[i].eof = 1;
			}
			else {
				eof = 0;
			}
		}
	}

	return (eof);
}

/* compare current characters. returns 0 if all characters are equal */
static
int fcmp_test_char (void)
{
	unsigned long i;

	for (i = 1; i < par_file_cnt; i++) {
		if (par_files[0].eof != par_files[i].eof) {
			return (1);
		}

		if (par_files[i].eof == 0) {
			if (par_files[0].chr != par_files[i].chr) {
				return (1);
			}
		}
	}

	return (0);
}

static
void fcmp_prt_char (FILE *fp, unsigned long ofs)
{
	unsigned long i;

	fprintf (fp, "%08lX", ofs);

	for (i = 0; i < par_file_cnt; i++) {
		if (par_files[i].eof) {
			fputs (" EOF", fp);
		}
		else {
			fprintf (fp, "  %02X", (unsigned) par_files[i].chr);
		}
	}

	fputs ("\n", fp);

	fflush (fp);
}

static
void fcmp_prt_header (FILE *fp)
{
	unsigned i;

	for (i = 0; i < par_file_cnt; i++) {
		if (par_files[i].fname == NULL) {
			fprintf (fp, "# %u: <stdin>\n", i + 1);
		}
		else {
#ifdef FCMP_STAT
			struct stat st;
			struct tm   *tm;

			if (stat (par_files[i].fname, &st)) {
				fprintf (fp, "# %u: %s\n", i + 1, par_files[i].fname);
			}
			else {
				tm = localtime (&st.st_mtime);

				fprintf (fp, "# %u: %s %04d-%02d-%02d %02d:%02d:%02d %lu\n", i + 1,
					par_files[i].fname,
					1900 + tm->tm_year, tm->tm_mon + 1, tm->tm_mday,
					tm->tm_hour, tm->tm_min, tm->tm_sec,
					(unsigned long) st.st_size
				);
			}
#else
		fprintf (fp, "# %u: %s\n", i + 1, par_files[i].fname);
#endif
		}
	}

	fflush (fp);
}

/* compare files. returns 0 if all files equal */
static
int fcmp_compare (void)
{
	int           equ;
	unsigned long cnt;
	unsigned long ofs;

	if (par_header) {
		fcmp_prt_header (stdout);
	}

	equ = 1;
	cnt = 0;
	ofs = 0;

	while (1) {
		if (fcmp_next_char()) {
			return (!equ);
		}

		if (fcmp_test_char()) {
			equ = 0;

			if (!par_quiet) {
				fcmp_prt_char (stdout, ofs);
			}

			cnt += 1;

			if ((par_diff_max > 0) && (cnt >= par_diff_max)) {
				return (!equ);
			}
		}

		ofs += 1;
	}

	return (!equ);
}

int main (int argc, char **argv)
{
	int i, r;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("fcmp");
			return (EXIT_SUCCESS);
		}
	}

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "q", "quiet")) {
			par_quiet = 1;
		}
		else if (str_isarg2 (argv[i], "h", "header")) {
			par_header = 1;
		}
		else if (str_isarg2 (argv[i], "f", "first")) {
			par_diff_max = 1;
		}
		else if (str_isarg2 (argv[i], "n", "max-diff")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing max diff count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_diff_max)) {
				fprintf (stderr, "%s: bad max diff count (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg1 (argv[i], "-")) {
			if (fcmp_new_file ("-")) {
				fprintf (stderr, "%s: can't open input file (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (argv[i][0] == '-') {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}
		else {
			if (fcmp_new_file (argv[i])) {
				fprintf (stderr, "%s: can't open input file (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}

		i += 1;
	}

	r = fcmp_compare();

	fcmp_close();

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
