/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/ihex/ihex.c                                              *
 * Created:     2004-06-09 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2004-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include "ihex.h"


static const char    *argv0 = NULL;

static int           par_enc = 1;
static int           par_list = 0;
static unsigned long par_addr = 0;
static unsigned      par_recsize = 16;
static char          *par_fname1 = NULL;
static char          *par_fname2 = NULL;
static int           par_drop = 0;


static
void prt_help (void)
{
	fputs (
		"ihex: encode, decode or list intel ihex files\n"
		"\n"
		"usage: ihex [options]\n"
		"  -a, --address int      Set the address for the next input file [0]\n"
		"  -c, --ihex string      Set an ihex input file name\n"
		"  -d, --decode           Decode from intel hex to binary\n"
		"  -e, --encode           Encode from binary to intel hex [default]\n"
		"  -f, --file int string  Set an input file name with address\n"
		"  -i string              Set input file name [stdin]\n"
		"  -l, --list             List the contents of an ihex file [no]\n"
		"  -n, --record-size int  Set the record size in bytes [16]\n"
		"  -o string              Set output file name [stdout]\n"
		"  -s, --segment int      Set segment address for encoding [0]\n",
		stdout
	);

	fflush (stdout);
}

static
int ihex_skip_line (FILE *fp)
{
	int c;

	while (1) {
		c = fgetc (fp);

		if (c == EOF) {
			return (1);
		}

		if ((c == 0x0a) || (c == 0x0d)) {
			return (0);
		}
	}

	return (0);
}

static
int ihex_get_hex8 (FILE *fp, unsigned char *val)
{
	unsigned i;
	int      c;

	*val = 0;

	for (i = 0; i < 2; i++) {
		c = fgetc (fp);

		if ((c >= '0') && (c <= '9')) {
			*val = (*val << 4) | (c - '0');
		}
		else if ((c >= 'A') && (c <= 'F')) {
			*val = (*val << 4) | (c - 'A' + 10);
		}
		else if ((c >= 'a') && (c <= 'f')) {
			*val = (*val << 4) | (c - 'a' + 10);
		}
		else {
			return (1);
		}
	}

	return (0);
}

static
void ihex_set_hex8 (FILE *fp, unsigned char c)
{
	int tmp;

	tmp = (c >> 4) & 0x0f;
	tmp += (tmp <= 9) ? '0' : ('A' - 10);
	fputc (tmp, fp);

	tmp = c & 0x0f;
	tmp += (tmp <= 9) ? '0' : ('A' - 10);
	fputc (tmp, fp);
}

static
unsigned char ihex_get_cksum (record_t *rec)
{
	unsigned      i;
	unsigned char ck;

	ck = rec->cnt & 0xff;
	ck += rec->addr & 0xff;
	ck += (rec->addr >> 8) & 0xff;
	ck += rec->type & 0xff;

	for (i = 0; i < rec->cnt; i++) {
		ck += rec->data[i];
	}

	ck = (~ck + 1) & 0xff;

	return (ck);
}

static
void ihex_init_record (record_t *rec, unsigned type, unsigned cnt, unsigned addr)
{
	rec->type = type;
	rec->cnt = cnt;
	rec->addr = addr;
}

static
int ihex_get_record (FILE *fp, record_t *rec)
{
	unsigned      i;
	int           c;
	unsigned char a1, a2;

	while (1) {
		c = fgetc (fp);

		if (c == EOF) {
			return (1);
		}

		if (c == ':') {
			break;
		}

		if ((c != 0x0a) && (c != 0x0d)) {
			ihex_skip_line (fp);
		}
	}

	if (ihex_get_hex8 (fp, &rec->cnt)) {
		return (1);
	}

	if (ihex_get_hex8 (fp, &a1) || ihex_get_hex8 (fp, &a2)) {
		return (1);
	}

	rec->addr = (a1 << 8) | a2;

	if (ihex_get_hex8 (fp, &rec->type)) {
		return (1);
	}

	for (i = 0; i < rec->cnt; i++) {
		if (ihex_get_hex8 (fp, &rec->data[i])) {
			return (1);
		}
	}

	if (ihex_get_hex8 (fp, &rec->cksum)) {
		return (1);
	}

	ihex_skip_line (fp);

	return (0);
}

static
void ihex_set_record (FILE *fp, record_t *rec)
{
	unsigned i;

	rec->cksum = ihex_get_cksum (rec);

	fprintf (fp, ":%02X%04X%02X",
		(unsigned) rec->cnt, (unsigned) rec->addr & 0xffffU, (unsigned) rec->type
	);

	for (i = 0; i < rec->cnt; i++) {
		ihex_set_hex8 (fp, rec->data[i]);
	}

	ihex_set_hex8 (fp, rec->cksum);

	fputs ("\n", fp);
}

static
void ihex_set_ulba (FILE *fp, unsigned long addr)
{
	record_t rec;

	ihex_init_record (&rec, IHEX_REC_ELAR, 2, 0);

	rec.data[0] = (addr >> 24) & 0xff;
	rec.data[1] = (addr >> 16) & 0xff;

	ihex_set_record (fp, &rec);
}

static
void ihex_set_usba (FILE *fp, unsigned short segm)
{
	record_t rec;

	ihex_init_record (&rec, IHEX_REC_ESAR, 2, 0);

	rec.data[0] = (segm >> 8) & 0xff;
	rec.data[1] = segm & 0xff;

	ihex_set_record (fp, &rec);
}

static
void ihex_set_end (FILE *fp)
{
	record_t rec;

	ihex_init_record (&rec, IHEX_REC_EOFR, 0, 0);

	ihex_set_record (fp, &rec);
}

static
FILE *ihex_open_next (unsigned long addr)
{
	unsigned i;
	char     dig;
	FILE     *fp;

	if (par_fname1 == NULL) {
		par_fname1 = "########.bin";
	}

	if (par_fname2 == NULL) {
		par_fname2 = malloc (strlen (par_fname1) + 1);
	}

	i = 0;
	while (par_fname1[i] != 0) {
		par_fname2[i] = par_fname1[i];
		i += 1;
	}

	par_fname2[i] = 0;

	while (i > 0) {
		i -= 1;

		if (par_fname2[i] == '#') {
			dig = addr & 0x0f;
			dig = (dig <= 9) ? ('0' + dig) : ('A' + dig - 10);
			par_fname2[i] = dig;
			addr = (addr >> 4);
		}
	}

	fp = fopen (par_fname2, "wb");

	return (fp);
}

static
void ihex_print_block (FILE *fp, unsigned long addr,
	unsigned seg, unsigned ofs, unsigned long cnt)
{
	fprintf (fp, "%08lX %04X:%04X %05lX\n", addr, seg, ofs, cnt);
}

static
int ihex_decode_fp (FILE *inp, int list)
{
	int           r;
	unsigned long start_addr, start_ulba, start_offs, start_cnt;
	unsigned long addr, raddr, ulba;
	record_t      rec;
	FILE          *fp;

	r = 1;
	addr = 0;
	ulba = 0;
	fp = NULL;

	start_addr = 0;
	start_ulba = 0;
	start_offs = 0;
	start_cnt = 0;

	while (1) {
		if (ihex_get_record (inp, &rec)) {
			break;
		}

		if (rec.cksum != ihex_get_cksum (&rec)) {
			fprintf (stderr, "%s: bad checksum T=%02X L=%04X CK=%02X/%02X",
				argv0,
				(unsigned) rec.type & 0xff, (unsigned) rec.addr & 0xffffU,
				(unsigned) rec.cksum, (unsigned) ihex_get_cksum (&rec)
			);

			if (par_drop) {
				fprintf (stderr, " (skipping)\n");
				continue;
			}

			fprintf (stderr, " (using anyway)\n");
		}

		if (rec.type == IHEX_REC_EOFR) {
			r = 0;
			break;
		}
		else if (rec.type == IHEX_REC_ELAR) {
			if (rec.cnt == 2) {
				ulba = (rec.data[0] << 8) | rec.data[1];
				ulba = ulba << 16;
			}
			else {
				fprintf (stderr,
					"%s: bad extended linear address record (ignoring)\n",
					argv0
				);
			}
		}
		else if (rec.type == IHEX_REC_ESAR) {
			if (rec.cnt == 2) {
				ulba = (rec.data[0] << 8) | rec.data[1];
				ulba = ulba << 4;
			}
			else {
				fprintf (stderr,
					"%s: bad extended segment address record (ignoring)\n",
					argv0
				);
			}
		}
		else if (rec.type == IHEX_REC_DATA) {
			raddr = ulba + (rec.addr & 0xffff);

			if (list) {
				if ((raddr == addr) && (start_cnt > 0)) {
					start_cnt += rec.cnt;
				}
				else {
					if (start_cnt > 0) {
						ihex_print_block (
							stdout,
							start_addr,
							(start_ulba >> 4) & 0xffff,
							start_offs & 0xffff,
							start_cnt
						);
					}

					start_addr = raddr;
					start_ulba = ulba;
					start_offs = rec.addr & 0xffff;
					start_cnt = rec.cnt;
				}
			}
			else {
				if ((fp != NULL) && (raddr != addr)) {
					fclose (fp);
					fp = NULL;
				}

				if (fp == NULL) {
					fp = ihex_open_next (raddr);
					if (fp == NULL) {
						break;
					}
				}

				if (fwrite (rec.data, 1, rec.cnt, fp) != rec.cnt) {
					break;
				}
			}

			addr = (raddr + rec.cnt) & 0xffffffff;
		}
		else {
			fprintf (stderr, "%s: unknown record type %02X at %04X (ignoring)\n",
				argv0, (unsigned) rec.type, (unsigned) rec.addr
			);
		}
	}

	if (list && (start_cnt > 0)) {
		ihex_print_block (
			stdout,
			start_addr,
			(start_ulba >> 4) & 0xffff,
			start_offs & 0xffff,
			start_cnt
		);
	}

	if (fp != NULL) {
		fclose (fp);
	}

	return (r);
}

static
int ihex_decode (const char *fname, int list)
{
	int  r;
	FILE *fp;

	if (fname == NULL) {
		r = ihex_decode_fp (stdin, list);
	}
	else {
		fp = fopen (fname, "rb");
		if (fp == NULL) {
			return (1);
		}

		r = ihex_decode_fp (fp, list);

		fclose (fp);
	}

	return (r);
}

static
int ihex_encode_ihex_fp (FILE *out, FILE *inp)
{
	record_t rec;

	while (1) {
		if (ihex_get_record (inp, &rec)) {
			break;
		}

		if (rec.type == IHEX_REC_EOFR) {
			break;
		}

		ihex_set_record (out, &rec);
	}

	return (0);
}

static
int ihex_encode_ihex (FILE *out, const char *inp)
{
	int  r;
	FILE *fp;

	fp = fopen (inp, "rb");
	if (fp == NULL) {
		return (1);
	}

	r = ihex_encode_ihex_fp (out, fp);

	fclose (fp);

	return (r);
}

static
int ihex_encode_fp (FILE *out, FILE *inp, unsigned long *addr, unsigned recsize)
{
	unsigned long ulba;
	record_t      rec;

	ulba = 0;

	while (1) {
		rec.type = 0x00;
		rec.addr = *addr & 0xffff;
		rec.cnt = fread (rec.data, 1, recsize, inp);

		if (rec.cnt == 0) {
			break;
		}

		if ((ulba & 0xffff0000) != (*addr & 0xffff0000)) {
			ihex_set_ulba (out, *addr);
			ulba = *addr & 0xffff0000;
		}

		ihex_set_record (out, &rec);

		*addr += rec.cnt;
	}

	return (0);
}

static
int ihex_encode (FILE *out, const char *inp, unsigned long *addr, unsigned recsize)
{
	int  r;
	FILE *fp;

	fp = fopen (inp, "rb");
	if (fp == NULL) {
		return (1);
	}

	r = ihex_encode_fp (out, fp, addr, recsize);

	fclose (fp);

	return (r);
}

int main (int argc, char *argv[])
{
	int  i;
	char *inp;
	FILE *out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("ihex");
			return (EXIT_SUCCESS);
		}
	}

	inp = NULL;
	out = stdout;

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "i", "input")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			if (par_enc) {
				if (ihex_encode (out, argv[i], &par_addr, par_recsize)) {
					return (EXIT_FAILURE);
				}
			}
			else {
				inp = argv[i];
			}
		}
		else if (str_isarg2 (argv[i], "o", "output")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			if (par_enc) {
				if (out != stdout) {
					fclose (out);
				}

				out = fopen (argv[i], "w");
				if (out == NULL) {
					return (EXIT_FAILURE);
				}
			}
			else {
				par_fname1 = argv[i];
				free (par_fname2);
				par_fname2 = NULL;
			}
		}
		else if (str_isarg2 (argv[i], "c", "ihex")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			if (par_enc) {
				if (ihex_encode_ihex (out, argv[i])) {
					return (EXIT_FAILURE);
				}
			}
			else {
				fprintf (stderr, "%s: invalid option (-c)\n", argv0);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "f", "file")) {
			if ((i + 2) >= argc) {
				fprintf (stderr, "%s: missing address or file name\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i + 1], &par_addr)) {
				fprintf (stderr, "%s: bad address (%s)\n", argv0, argv[i + 1]);
				return (EXIT_FAILURE);
			}

			if (par_enc) {
				if (ihex_encode (out, argv[i + 2], &par_addr, par_recsize)) {
					return (EXIT_FAILURE);
				}
			}
			else {
				inp = argv[i];
			}

			i += 2;
		}
		else if (str_isarg2 (argv[i], "e", "encode")) {
			par_enc = 1;
		}
		else if (str_isarg2 (argv[i], "d", "decode")) {
			par_enc = 0;
		}
		else if (str_isarg2 (argv[i], "l", "list")) {
			par_enc = 0;
			par_list = 1;
		}
		else if (str_isarg2 (argv[i], "n", "record-size")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing record size\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_recsize)) {
				fprintf (stderr, "%s: bad record size (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			if (par_recsize < 1) {
				par_recsize = 1;
			}
		}
		else if (str_isarg2 (argv[i], "a", "address")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing address\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_addr)) {
				fprintf (stderr, "%s: bad address (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "s", "segment")) {
			unsigned long segm;

			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing segment address\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &segm)) {
				fprintf (stderr, "%s: bad segment address (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			if (par_enc) {
				ihex_set_usba (out, segm & 0xffff);
				par_addr = 0;
			}
		}
		else if (argv[i][0] != '-') {
			if (par_enc) {
				if (ihex_encode (out, argv[i], &par_addr, par_recsize)) {
					return (EXIT_FAILURE);
				}
			}
			else {
				if (inp == NULL) {
					inp = argv[i];
				}
				else if (par_fname1 == NULL) {
					par_fname1 = argv[i];
					free (par_fname2);
					par_fname2 = NULL;
				}
				else {
					return (EXIT_FAILURE);
				}
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (par_enc) {
		ihex_set_end (out);

		if (out != stdout) {
			fclose (out);
		}
	}
	else {
		if (ihex_decode (inp, par_list)) {
			return (EXIT_FAILURE);
		}
	}

	return (EXIT_SUCCESS);
}
