/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/socket.c                                             *
 * Created:     2002-01-20 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by  the Free Software Foundation.                                         *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#define _BSD_SOURCE 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "socket.h"


static
char *tcp_str_clone (const char *src)
{
	unsigned long i, n;
	char          *dst;

	n = 0;
	while (src[n] != 0) {
		n += 1;
	}

	n += 1;

	dst = (char *) malloc (n);

	if (dst == NULL) {
		return (NULL);
	}

	for (i = 0; i < n; i++) {
		dst[i] = src[i];
	}

	return (dst);
}

int tcp_connect (const char *host, unsigned short port)
{
	int                fd;
	struct sockaddr_in addr;

	fd = socket (PF_INET, SOCK_STREAM, 0);
	if (fd < 0) {
		return (-1);
	}

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr (host);
	addr.sin_port = htons (port);

	if (connect (fd, (struct sockaddr *) &addr, sizeof (addr)) == -1) {
		close (fd);
		return (-1);
	}

	return (fd);
}

int tcp_listen (const char *host, unsigned short port, unsigned queue)
{
	int                fd;
	struct sockaddr_in addr;

	addr.sin_family = AF_INET;

	if ((host == NULL) || (host[0] == 0)) {
		addr.sin_addr.s_addr = INADDR_ANY;
	}
	else {
		if (inet_aton (host, &addr.sin_addr) == 0) {
			return (-1);
		}
	}

#if USHRT_MAX != 65535
	if (port > 65535) {
		return (-1);
	}
#endif

	addr.sin_port = htons ((unsigned short) port);

	fd = socket (PF_INET, SOCK_STREAM, 0);
	if (fd < 0) {
		return (-1);
	}

	if (bind (fd, (struct sockaddr *) &addr, sizeof (addr)) == -1) {
		close (fd);
		return (-1);
	}

	if (listen (fd, (int) queue) == -1) {
		close (fd);
		return (-1);
	}

	return (fd);
}

int tcp_accept (int fd)
{
	int                ret;
	struct sockaddr_in addr;
	socklen_t          cnt;

	cnt = sizeof (addr);
	ret = accept (fd, (struct sockaddr *) &addr, &cnt);

	return (ret);
}

int tcp_accept_time (int fd, int t)
{
	int                r;
	struct pollfd      pfd[1];
	struct sockaddr_in addr;
	socklen_t          cnt;

	pfd[0].fd = fd;
	pfd[0].events = POLLIN;

	r = poll (pfd, 1, t);
	if (r < 0) {
		return (-1);
	}

	if ((pfd[0].revents & POLLIN) == 0) {
		return (-1);
	}

	cnt = sizeof (addr);
	r = accept (fd, (struct sockaddr *) &addr, &cnt);

	return (r);
}

void tcp_shutdown (int fd)
{
	shutdown (fd, SHUT_RDWR);
}

void tcp_shutdown_out (int fd)
{
	shutdown (fd, SHUT_WR);
}

void tcp_shutdown_inp (int fd)
{
	shutdown (fd, SHUT_RD);
}

int tcp_get_peer (int fd, char **host, unsigned short *port)
{
	struct sockaddr_in addr;
	socklen_t          cnt;

	cnt = (socklen_t) sizeof (addr);
	if (getpeername (fd, (struct sockaddr *) &addr, &cnt) == -1) {
		return (1);
	}

	*host = tcp_str_clone (inet_ntoa (addr.sin_addr));
	*port = (unsigned short) ntohs (addr.sin_port);

	return (0);
}

int tcp_get_host (int fd, char **host, unsigned short *port)
{
	struct sockaddr_in addr;
	socklen_t          cnt;

	cnt = (socklen_t) sizeof (addr);
	if (getsockname (fd, (struct sockaddr *) &addr, &cnt) == -1) {
		return (1);
	}

	*host = tcp_str_clone (inet_ntoa (addr.sin_addr));
	*port = (unsigned) ntohs (addr.sin_port);

	return (0);
}

char *tcp_addr_get_host (const char *str)
{
	unsigned i, n;
	char     *ret;

	n = 0;
	while ((str[n] != 0) && (str[n] != ':')) {
		n += 1;
	}

	ret = (char *) malloc (n + 1);
	if (ret == NULL) {
		return (NULL);
	}

	for (i = 0; i < n; i++) {
		ret[i] = str[i];
	}

	ret[n] = 0;

	return (ret);
}

int tcp_addr_get_port (const char *str, unsigned short *port)
{
	unsigned i;

	i = 0;
	while ((str[i] != 0) && (str[i] != ':')) {
		i += 1;
	}

	if (str[i] != ':') {
		return (1);
	}

	i += 1;

	if ((str[i] < '0') || (str[i] > '9')) {
		return (1);
	}

	*port = (unsigned short) (str[i] - '0');

	i += 1;

	while ((str[i] >= '0') && (str[i] <= '9')) {
		*port = 10 * *port + (unsigned short) (str[i] - '0');
		i += 1;
	}

	return (0);
}

char *tcp_host_resolve (const char *str)
{
	char           buf[256];
	struct hostent *ent;
	char           *ret;

	ent = gethostbyname (str);
	if (ent == NULL) {
		return (NULL);
	}

	if (inet_ntop (ent->h_addrtype, ent->h_addr_list[0], buf, 256) == NULL) {
		return (NULL);
	}

	ret = tcp_str_clone (buf);

	return (ret);
}
