/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lnconv/lnconv.c                                          *
 * Created:     2000-09-22 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2000-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/stresc.h>
#include <lib/strtoint.h>

#include "lnconv.h"


static const char    *argv0 = NULL;

static int           par_do_lnconv = 0;
static int           par_do_rtrim = 0;

static unsigned char par_eol[LNCONV_STR_MAX];
static unsigned      par_eol_cnt = 0;

static unsigned char par_rtrim[LNCONV_STR_MAX];
static unsigned      par_rtrim_cnt = 0;


static
void prt_help (void)
{
	fputs (
		"Usage: lnconv [options] [input [output]]\n"
		"  -i string         Set input file name [stdin]\n"
		"  -o string         Set output file name [stdout]\n"
		"  -e, --eol string  Set end of line type [unix]\n"
		"  -s string         Set literal end of line string\n"
		"  -r, --rtrim       Remove white space from end of lines [no]\n"
		"  -R string         Remove characters in string from end of lines\n"
		"Recognized end of line types are:\n"
		"  lf, unix   (LF)\n"
		"  crlf, dos  (CR+LF)\n"
		"  cr, mac    (CR)\n"
		"  null       (0)\n",
		stdout
	);

	fflush (stdout);
}

static
void line_init (line_t *ln)
{
	ln->i = 0;
	ln->n = 0;
	ln->buf = NULL;

	ln->eol_cnt = 0;
}

static
void line_free (line_t *ln)
{
	free (ln->buf);
}

static
int line_add_char (line_t *ln, unsigned char c)
{
	if (ln->i >= ln->n) {
		unsigned      n;
		unsigned char *tmp;

		n = ln->i;
		while (n <= ln->n) {
			n = 2 * n + 1;
		}

		tmp = realloc (ln->buf, n);
		if (tmp == NULL) {
			return (1);
		}

		ln->buf = tmp;
		ln->n = n;
	}

	ln->buf[ln->i] = c;
	ln->i += 1;

	return (0);
}

static
int line_read (line_t *ln, FILE *fp)
{
	int c;

	ln->i = 0;
	ln->eol_cnt = 0;

	c = fgetc (fp);
	while (c != EOF) {
		switch (c) {
			case 0x0d:
				ln->eol[0] = 0x0d;
				ln->eol_cnt = 1;

				c = fgetc (fp);
				if (c == 0x0a) {
					ln->eol[1] = 0x0a;
					ln->eol_cnt = 2;
				}
				else {
					ungetc (c, fp);
				}
				return (0);

			case 0x0a:
				ln->eol[0] = 0x0a;
				ln->eol_cnt = 1;
				return (0);

			default:
				if (line_add_char (ln, c)) {
					return (1);
				}
				break;
		}

		c = fgetc (fp);
	}

	return (0);
}

static
int line_write (line_t *ln, FILE *fp)
{
	if (ln->i > 0) {
		if (fwrite (ln->buf, 1, ln->i, fp) != ln->i) {
			return (1);
		}
	}

	if (ln->eol_cnt > 0) {
		if (fwrite (ln->eol, 1, ln->eol_cnt, fp) != ln->eol_cnt) {
			return (1);
		}
	}

	return (0);
}

static
int line_set_eol (line_t *ln, unsigned char *buf, unsigned n)
{
	if (n > LNCONV_STR_MAX) {
		return (1);
	}

	memcpy (ln->eol, buf, n);
	ln->eol_cnt = n;

	return (0);
}

static
int char_in_set (unsigned char c, const unsigned char *buf, unsigned n)
{
	while (n > 0) {
		if (*buf == c) {
			return (1);
		}

		buf += 1;
		n -= 1;
	}

	return (0);
}

static
void line_rtrim (line_t *ln, unsigned char *buf, unsigned n)
{
	while (ln->i > 0) {
		if (char_in_set (ln->buf[ln->i - 1], buf, n)) {
			ln->i -= 1;
		}
		else {
			return;
		}
	}
}

static
int line_is_eof (line_t *ln)
{
	if ((ln->i == 0) && (ln->eol_cnt == 0)) {
		return (1);
	}

	return (0);
}

static
int lnconv (FILE *inp, FILE *out)
{
	line_t ln;

	line_init (&ln);

	while (line_read (&ln, inp) == 0) {
		if (line_is_eof (&ln)) {
			break;
		}

		if (par_do_lnconv) {
			line_set_eol (&ln, par_eol, par_eol_cnt);
		}

		if (par_do_rtrim) {
			line_rtrim (&ln, par_rtrim, par_rtrim_cnt);
		}

		if (line_write (&ln, out)) {
			line_free (&ln);
			return (1);
		}
	}

	line_free (&ln);

	return (0);
}

int main (int argc, char *argv[])
{
	int      i, r;
	unsigned fcnt;
	char     *finp, *fout;
	FILE     *inp, *out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("lnconv");
			return (EXIT_SUCCESS);
		}
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "e", "end-of-line")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing end of line type\n", argv0);
				return (EXIT_FAILURE);
			}

			par_do_lnconv = 1;

			if ((strcmp (argv[i], "lf") == 0) || (strcmp (argv[i], "unix") == 0)) {
				par_eol[0] = 0x0a;
				par_eol_cnt = 1;
			}
			else if ((strcmp (argv[i], "cr") == 0) || (strcmp (argv[i], "mac") == 0)) {
				par_eol[0] = 0x0d;
				par_eol_cnt = 1;
			}
			else if ((strcmp (argv[i], "crlf") == 0) || (strcmp (argv[i], "dos") == 0)) {
				par_eol[0] = 0x0d;
				par_eol[1] = 0x0a;
				par_eol_cnt = 2;
			}
			else if (strcmp (argv[i], "null") == 0) {
				par_eol[0] = 0x00;
				par_eol_cnt = 1;
			}
			else {
				fprintf (stderr, "%s: unknown end of line type (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg1 (argv[i], "s")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing end of line string\n", argv0);
				return (EXIT_FAILURE);
			}

			par_do_lnconv = 1;

			par_eol_cnt = buf_escape (par_eol, (unsigned char *) argv[i],
				LNCONV_STR_MAX
			);
		}
		else if (str_isarg2 (argv[i], "r", "rtrim")) {
			par_do_rtrim = 1;

			par_rtrim[0] = 0x20;
			par_rtrim[1] = 0x09;
			par_rtrim_cnt = 2;
		}
		else if (str_isarg1 (argv[i], "R")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing character set\n", argv0);
				return (EXIT_FAILURE);
			}

			par_do_rtrim = 1;
			par_rtrim_cnt = buf_escape (par_rtrim, (unsigned char *) argv[i],
				LNCONV_STR_MAX
			);
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	inp = (finp == NULL) ? stdin : fopen (finp, "rb");
	if (inp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
		return (EXIT_FAILURE);
	}

	out = (fout == NULL) ? stdout : fopen (fout, "wb");
	if (out == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
		return (EXIT_FAILURE);
	}

	r = lnconv (inp, out);

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
