/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mdig/mdig.c                                              *
 * Created:     2005-07-19 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2005-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <lib/adler32.h>
#include <lib/crc.h>
#include <lib/md2.h>
#include <lib/md4.h>
#include <lib/md5.h>
#include <lib/rmd160.h>
#include <lib/sha1.h>
#include <lib/sha224.h>
#include <lib/sha256.h>

/* sha384 and sha512 require 64bit integers */
#ifndef MISC_DOS
#include <lib/sha384.h>
#include <lib/sha512.h>
#endif

#include "mdig.h"


#define MDIG_BUF_SIZE MISC_BUF_SIZE


static const char *argv0 = NULL;

static int      par_bin = 0;
static int      par_pipe = 0;
static int      par_capital = 0;
static unsigned par_digits = 0;
static int      par_prefix = 0;

static mdig_t   *par_mdig = NULL;
static mdig_t   *par_mdig_tl = NULL;


static
void prt_help (void)
{
	fputs (
		"mdig: calculate message digests\n"
		"\n"
		"usage: mdig [options] [input [output]]\n"
		"  -i string            Set input file name [stdin]\n"
		"  -o string            Set output file name [stdout]\n"
		"  -b, --binary         Write the output as binary string [no]\n"
		"  -c, --capital        Use capital letters as hex digits [no]\n"
		"  -m, --digest string  Set a message digest to calculate [none]\n"
		"  -n, --digits int     Group n digits in output [0]\n"
		"  -p, --pipe           Pass input through to stdout [no]\n"
		"  -t, --prefix         Print the digest name in the output [no]\n"
		"\n"
		"Digests: size, sum8, adler8, adler16, adler32, crc, md2, md4, md5,\n"
		"         ed2k, rmd160, sha1, sha224, sha256"
#ifndef MISC_DOS
		", sha384, sha512"
#endif
		"\n",
		stdout
	);

	fflush (stdout);
}

static
mdig_t *mdig_new (void *ext)
{
	mdig_t *mdig;

	mdig = malloc (sizeof (mdig_t));
	if (mdig == NULL) {
		return (NULL);
	}

	mdig->next = NULL;
	mdig->del = NULL;
	mdig->calc = NULL;
	mdig->get = NULL;
	mdig->ext = ext;

	return (mdig);
}


/*****************************************************************************
 * size
 *****************************************************************************/

static
void mdig_size_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_size_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	unsigned long long *ext;

	ext = mdig->ext;

	*ext += n;
}

static
unsigned mdig_size_get (mdig_t *mdig, void *buf, unsigned max)
{
	unsigned           i, j, n;
	unsigned char      c;
	unsigned long long val;
	unsigned char      *tmp;

	tmp = buf;
	val = *(unsigned long long *) mdig->ext;

	n = 0;
	while (n < max) {
		tmp[n] = val & 0xff;
		val = val >> 8;

		n += 1;

		if (val == 0) {
			break;
		}
	}

	if ((val != 0) || (n == 0)) {
		return (0);
	}

	i = 0;
	j = n - 1;
	while (i < j) {
		c = tmp[i];
		tmp[i] = tmp[j];
		tmp[j] = c;

		i += 1;
		j -= 1;
	}

	return (n);
}

static
mdig_t *mdig_size_new (void)
{
	unsigned long long *ext;
	mdig_t             *mdig;

	ext = malloc (sizeof (unsigned long long));
	if (ext == NULL) {
		return (NULL);
	}

	*ext = 0;

	mdig = mdig_new (ext);
	if (mdig == NULL) {
		free (ext);
		return (NULL);
	}

	mdig->name = "size";
	mdig->del = mdig_size_del;
	mdig->calc = mdig_size_calc;
	mdig->get = mdig_size_get;

	return (mdig);
}


/*****************************************************************************
 * sum-8
 *****************************************************************************/

static
void mdig_sum8_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_sum8_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	unsigned char       *ext;
	const unsigned char *tmp;

	ext = mdig->ext;
	tmp = buf;

	while (n > 0) {
		*ext += *(tmp++);
		n -= 1;
	}

	*ext &= 0xff;
}

static
unsigned mdig_sum8_get (mdig_t *mdig, void *buf, unsigned max)
{
	unsigned char *ext;
	unsigned char *tmp;

	if (max < 1) {
		return (0);
	}

	ext = mdig->ext;
	tmp = buf;

	tmp[0] = *ext & 0xff;

	return (1);
}

static
mdig_t *mdig_sum8_new (void)
{
	unsigned char *ext;
	mdig_t        *mdig;

	ext = malloc (1);
	if (ext == NULL) {
		return (NULL);
	}

	*ext = 0;

	mdig = mdig_new (ext);
	if (mdig == NULL) {
		free (ext);
		return (NULL);
	}

	mdig->name = "sum8";
	mdig->del = mdig_sum8_del;
	mdig->calc = mdig_sum8_calc;
	mdig->get = mdig_sum8_get;

	return (mdig);
}


/*****************************************************************************
 * adler-8
 *****************************************************************************/

static
void mdig_adler8_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_adler8_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	adler8_calc (mdig->ext, buf, n);
}

static
unsigned mdig_adler8_get (mdig_t *mdig, void *buf, unsigned max)
{
	unsigned      val;
	unsigned char *tmp = buf;

	if (max < 1) {
		return (0);
	}

	adler8_done (mdig->ext);

	val = adler8_get_sum (mdig->ext);

	tmp[0] = val & 0xff;

	return (1);
}

static
mdig_t *mdig_adler8_new (void)
{
	adler8_t *ad;
	mdig_t   *mdig;

	ad = malloc (sizeof (adler8_t));
	if (ad == NULL) {
		return (NULL);
	}

	adler8_init (ad);

	mdig = mdig_new (ad);
	if (mdig == NULL) {
		free (ad);
		return (NULL);
	}

	mdig->name = "adler8";
	mdig->del = mdig_adler8_del;
	mdig->calc = mdig_adler8_calc;
	mdig->get = mdig_adler8_get;

	return (mdig);
}


/*****************************************************************************
 * adler-16
 *****************************************************************************/

static
void mdig_adler16_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_adler16_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	adler16_calc (mdig->ext, buf, n);
}

static
unsigned mdig_adler16_get (mdig_t *mdig, void *buf, unsigned max)
{
	unsigned      val;
	unsigned char *tmp = buf;

	if (max < 2) {
		return (0);
	}

	adler16_done (mdig->ext);

	val = adler16_get_sum (mdig->ext);

	tmp[0] = (val >> 8) & 0xff;
	tmp[1] = val & 0xff;

	return (2);
}

static
mdig_t *mdig_adler16_new (void)
{
	adler16_t *ad;
	mdig_t    *mdig;

	ad = malloc (sizeof (adler16_t));
	if (ad == NULL) {
		return (NULL);
	}

	adler16_init (ad);

	mdig = mdig_new (ad);
	if (mdig == NULL) {
		free (ad);
		return (NULL);
	}

	mdig->name = "adler16";
	mdig->del = mdig_adler16_del;
	mdig->calc = mdig_adler16_calc;
	mdig->get = mdig_adler16_get;

	return (mdig);
}


/*****************************************************************************
 * adler-32
 *****************************************************************************/

static
void mdig_adler32_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_adler32_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	adler32_calc (mdig->ext, buf, n);
}

static
unsigned mdig_adler32_get (mdig_t *mdig, void *buf, unsigned max)
{
	unsigned long val;
	unsigned char *tmp = buf;

	if (max < 4) {
		return (0);
	}

	adler32_done (mdig->ext);

	val = adler32_get_sum (mdig->ext);

	tmp[0] = (val >> 24) & 0xff;
	tmp[1] = (val >> 16) & 0xff;
	tmp[2] = (val >> 8) & 0xff;
	tmp[3] = val & 0xff;

	return (4);
}

static
mdig_t *mdig_adler32_new (void)
{
	adler32_t *ad;
	mdig_t    *mdig;

	ad = malloc (sizeof (adler32_t));
	if (ad == NULL) {
		return (NULL);
	}

	adler32_init (ad);

	mdig = mdig_new (ad);
	if (mdig == NULL) {
		free (ad);
		return (NULL);
	}

	mdig->name = "adler32";
	mdig->del = mdig_adler32_del;
	mdig->calc = mdig_adler32_calc;
	mdig->get = mdig_adler32_get;

	return (mdig);
}


/*****************************************************************************
 * crc
 *****************************************************************************/

static
void mdig_crc_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_crc_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	crc32_calc (mdig->ext, buf, n);
}

static
unsigned mdig_crc_get (mdig_t *mdig, void *buf, unsigned max)
{
	unsigned long val;
	unsigned char *tmp = buf;

	if (max < 4) {
		return (0);
	}

	crc32_done (mdig->ext);

	val = crc32_get_crc (mdig->ext);

	tmp[0] = (val >> 24) & 0xff;
	tmp[1] = (val >> 16) & 0xff;
	tmp[2] = (val >> 8) & 0xff;
	tmp[3] = val & 0xff;

	return (4);
}

static
mdig_t *mdig_crc_new (void)
{
	crc32_t *crc;
	mdig_t  *mdig;

	crc = malloc (sizeof (crc32_t));
	if (crc == NULL) {
		return (NULL);
	}

	crc32_init (crc, CRC_CRC32, CRC_POSIX);

	mdig = mdig_new (crc);
	if (mdig == NULL) {
		free (crc);
		return (NULL);
	}

	mdig->name = "crc";
	mdig->del = mdig_crc_del;
	mdig->calc = mdig_crc_calc;
	mdig->get = mdig_crc_get;

	return (mdig);
}


/*****************************************************************************
 * md2
 *****************************************************************************/

static
void mdig_md2_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_md2_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	md2_calc (mdig->ext, buf, n);
}

static
unsigned mdig_md2_get (mdig_t *mdig, void *buf, unsigned max)
{
	md2_done (mdig->ext);

	return (md2_get_digest (mdig->ext, buf, max));
}

static
mdig_t *mdig_md2_new (void)
{
	md2_t  *md2;
	mdig_t *mdig;

	md2 = malloc (sizeof (md2_t));
	if (md2 == NULL) {
		return (NULL);
	}

	md2_init (md2);

	mdig = mdig_new (md2);
	if (mdig == NULL) {
		free (md2);
		return (NULL);
	}

	mdig->name = "md2";
	mdig->del = mdig_md2_del;
	mdig->calc = mdig_md2_calc;
	mdig->get = mdig_md2_get;

	return (mdig);
}


/*****************************************************************************
 * md4
 *****************************************************************************/

static
void mdig_md4_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_md4_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	md4_calc (mdig->ext, buf, n);
}

static
unsigned mdig_md4_get (mdig_t *mdig, void *buf, unsigned max)
{
	if (max < 16) {
		return (0);
	}

	md4_done (mdig->ext);
	md4_get_sum (mdig->ext, buf);

	return (16);
}

static
mdig_t *mdig_md4_new (void)
{
	md4_t  *md4;
	mdig_t *mdig;

	md4 = malloc (sizeof (md4_t));
	if (md4 == NULL) {
		return (NULL);
	}

	md4_init (md4);

	mdig = mdig_new (md4);
	if (mdig == NULL) {
		free (md4);
		return (NULL);
	}

	mdig->name = "md4";
	mdig->del = mdig_md4_del;
	mdig->calc = mdig_md4_calc;
	mdig->get = mdig_md4_get;

	return (mdig);
}


/*****************************************************************************
 * md5
 *****************************************************************************/

static
void mdig_md5_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_md5_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	md5_calc (mdig->ext, buf, n);
}

static
unsigned mdig_md5_get (mdig_t *mdig, void *buf, unsigned max)
{
	if (max < 16) {
		return (0);
	}

	md5_done (mdig->ext);
	md5_get_sum (mdig->ext, buf);

	return (16);
}

static
mdig_t *mdig_md5_new (void)
{
	md5_t  *md5;
	mdig_t *mdig;

	md5 = malloc (sizeof (md5_t));
	if (md5 == NULL) {
		return (NULL);
	}

	md5_init (md5);

	mdig = mdig_new (md5);
	if (mdig == NULL) {
		free (md5);
		return (NULL);
	}

	mdig->name = "md5";
	mdig->del = mdig_md5_del;
	mdig->calc = mdig_md5_calc;
	mdig->get = mdig_md5_get;

	return (mdig);
}


/*****************************************************************************
 * ripemd-160
 *****************************************************************************/

static
void mdig_rmd160_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_rmd160_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	rmd160_calc (mdig->ext, buf, n);
}

static
unsigned mdig_rmd160_get (mdig_t *mdig, void *buf, unsigned max)
{
	if (max < 20) {
		return (0);
	}

	rmd160_done (mdig->ext);
	rmd160_get_sum (mdig->ext, buf);

	return (20);
}

static
mdig_t *mdig_rmd160_new (void)
{
	rmd160_t *rmd;
	mdig_t   *mdig;

	rmd = malloc (sizeof (rmd160_t));
	if (rmd == NULL) {
		return (NULL);
	}

	rmd160_init (rmd);

	mdig = mdig_new (rmd);
	if (mdig == NULL) {
		free (rmd);
		return (NULL);
	}

	mdig->name = "rmd160";
	mdig->del = mdig_rmd160_del;
	mdig->calc = mdig_rmd160_calc;
	mdig->get = mdig_rmd160_get;

	return (mdig);
}


/*****************************************************************************
 * sha-1
 *****************************************************************************/

static
void mdig_sha1_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_sha1_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	sha1_calc (mdig->ext, buf, n);
}

static
unsigned mdig_sha1_get (mdig_t *mdig, void *buf, unsigned max)
{
	if (max < 20) {
		return (0);
	}

	sha1_done (mdig->ext);
	sha1_get_sum (mdig->ext, buf);

	return (20);
}

static
mdig_t *mdig_sha1_new (void)
{
	sha1_t *sha1;
	mdig_t *mdig;

	sha1 = malloc (sizeof (sha1_t));
	if (sha1 == NULL) {
		return (NULL);
	}

	sha1_init (sha1);

	mdig = mdig_new (sha1);
	if (mdig == NULL) {
		free (sha1);
		return (NULL);
	}

	mdig->name = "sha1";
	mdig->del = mdig_sha1_del;
	mdig->calc = mdig_sha1_calc;
	mdig->get = mdig_sha1_get;

	return (mdig);
}


/*****************************************************************************
 * sha-224
 *****************************************************************************/

static
void mdig_sha224_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_sha224_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	sha224_calc (mdig->ext, buf, n);
}

static
unsigned mdig_sha224_get (mdig_t *mdig, void *buf, unsigned max)
{
	if (max < 28) {
		return (0);
	}

	sha224_done (mdig->ext);
	sha224_get_sum (mdig->ext, buf, max);

	return (28);
}

static
mdig_t *mdig_sha224_new (void)
{
	sha224_t *sha;
	mdig_t   *mdig;

	sha = malloc (sizeof (sha224_t));
	if (sha == NULL) {
		return (NULL);
	}

	sha224_init (sha);

	mdig = mdig_new (sha);
	if (mdig == NULL) {
		free (sha);
		return (NULL);
	}

	mdig->name = "sha224";
	mdig->del = mdig_sha224_del;
	mdig->calc = mdig_sha224_calc;
	mdig->get = mdig_sha224_get;

	return (mdig);
}


/*****************************************************************************
 * sha-256
 *****************************************************************************/

static
void mdig_sha256_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_sha256_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	sha256_calc (mdig->ext, buf, n);
}

static
unsigned mdig_sha256_get (mdig_t *mdig, void *buf, unsigned max)
{
	if (max < 32) {
		return (0);
	}

	sha256_done (mdig->ext);
	sha256_get_sum (mdig->ext, buf, max);

	return (32);
}

static
mdig_t *mdig_sha256_new (void)
{
	sha256_t *sha;
	mdig_t   *mdig;

	sha = malloc (sizeof (sha256_t));
	if (sha == NULL) {
		return (NULL);
	}

	sha256_init (sha);

	mdig = mdig_new (sha);
	if (mdig == NULL) {
		free (sha);
		return (NULL);
	}

	mdig->name = "sha256";
	mdig->del = mdig_sha256_del;
	mdig->calc = mdig_sha256_calc;
	mdig->get = mdig_sha256_get;

	return (mdig);
}


#ifndef MISC_DOS

/*****************************************************************************
 * sha-384
 *****************************************************************************/

static
void mdig_sha384_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_sha384_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	sha384_calc (mdig->ext, buf, n);
}

static
unsigned mdig_sha384_get (mdig_t *mdig, void *buf, unsigned max)
{
	if (max < 48) {
		return (0);
	}

	sha384_done (mdig->ext);
	sha384_get_sum (mdig->ext, buf, max);

	return (48);
}

static
mdig_t *mdig_sha384_new (void)
{
	sha384_t *sha;
	mdig_t   *mdig;

	sha = malloc (sizeof (sha384_t));
	if (sha == NULL) {
		return (NULL);
	}

	sha384_init (sha);

	mdig = mdig_new (sha);
	if (mdig == NULL) {
		free (sha);
		return (NULL);
	}

	mdig->name = "sha384";
	mdig->del = mdig_sha384_del;
	mdig->calc = mdig_sha384_calc;
	mdig->get = mdig_sha384_get;

	return (mdig);
}

#endif


#ifndef MISC_DOS

/*****************************************************************************
 * sha-512
 *****************************************************************************/

static
void mdig_sha512_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_sha512_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	sha512_calc (mdig->ext, buf, n);
}

static
unsigned mdig_sha512_get (mdig_t *mdig, void *buf, unsigned max)
{
	if (max < 64) {
		return (0);
	}

	sha512_done (mdig->ext);
	sha512_get_sum (mdig->ext, buf, max);

	return (64);
}

static
mdig_t *mdig_sha512_new (void)
{
	sha512_t *sha;
	mdig_t   *mdig;

	sha = malloc (sizeof (sha512_t));
	if (sha == NULL) {
		return (NULL);
	}

	sha512_init (sha);

	mdig = mdig_new (sha);
	if (mdig == NULL) {
		free (sha);
		return (NULL);
	}

	mdig->name = "sha512";
	mdig->del = mdig_sha512_del;
	mdig->calc = mdig_sha512_calc;
	mdig->get = mdig_sha512_get;

	return (mdig);
}

#endif


/*****************************************************************************
 * ed2k
 *****************************************************************************/

static
void mdig_ed2k_del (mdig_t *mdig)
{
	if (mdig != NULL) {
		free (mdig->ext);
		free (mdig);
	}
}

static
void mdig_ed2k_calc (mdig_t *mdig, const void *buf, unsigned long n)
{
	unsigned long       i;
	ed2k_t              *ed2k;
	const unsigned char *tmp;

	ed2k = mdig->ext;
	tmp = buf;

	while (n >= (ED2K_PARTSIZE - ed2k->psize)) {
		i = ED2K_PARTSIZE - ed2k->psize;

		md4_calc (&ed2k->hash1, buf, i);
		md4_done (&ed2k->hash1);
		md4_get_sum (&ed2k->hash1, ed2k->buf);
		md4_calc (&ed2k->hash2, ed2k->buf, 16);

		md4_init (&ed2k->hash1);

		tmp += i;
		n -= i;

		ed2k->psize = 0;
		ed2k->parts += 1;
	}

	if (n > 0) {
		md4_calc (&ed2k->hash1, tmp, n);
		ed2k->psize += n;
	}
}

static
unsigned mdig_ed2k_get (mdig_t *mdig, void *buf, unsigned max)
{
	ed2k_t *ed2k;

	if (max < 16) {
		return (0);
	}

	ed2k = mdig->ext;

	if ((ed2k->psize > 0) || (ed2k->parts == 0)) {
		md4_done (&ed2k->hash1);
		md4_get_sum (&ed2k->hash1, ed2k->buf);
		md4_calc (&ed2k->hash2, ed2k->buf, 16);
		md4_init (&ed2k->hash1);
		ed2k->parts += 1;
	}

	if (ed2k->parts > 1) {
		md4_done (&ed2k->hash2);
		md4_get_sum (&ed2k->hash2, buf);
		md4_init (&ed2k->hash2);
	}
	else {
		memcpy (buf, ed2k->buf, 16);
	}

	return (16);
}

static
mdig_t *mdig_ed2k_new (void)
{
	ed2k_t *ed2k;
	mdig_t *mdig;

	ed2k = malloc (sizeof (ed2k_t));
	if (ed2k == NULL) {
		return (NULL);
	}

	ed2k->parts = 0;
	ed2k->psize = 0;
	md4_init (&ed2k->hash1);
	md4_init (&ed2k->hash2);

	mdig = mdig_new (ed2k);
	if (mdig == NULL) {
		free (ed2k);
		return (NULL);
	}

	mdig->name = "ed2k";
	mdig->del = mdig_ed2k_del;
	mdig->calc = mdig_ed2k_calc;
	mdig->get = mdig_ed2k_get;

	return (mdig);
}


static
char mdig_get_hex (unsigned char val)
{
	if (val < 10) {
		return ((char) ('0' + val));
	}

	if (par_capital) {
		return ((char) ('A' + val - 10));
	}

	return ((char) ('a' + val - 10));
}

static
int mdig_calc (FILE *inp, FILE *out, mdig_t *mdig)
{
	unsigned      i, j;
	unsigned long n;
	mdig_t        *tmp;
	unsigned char buf[MDIG_BUF_SIZE + 256];

	n = fread (buf, 1, MDIG_BUF_SIZE, inp);

	while (n > 0) {
		tmp = mdig;

		while (tmp != NULL) {
			tmp->calc (tmp, buf, n);
			tmp = tmp->next;
		}

		if (par_pipe) {
			if (fwrite (buf, 1, n, stdout) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}

		n = fread (buf, 1, MDIG_BUF_SIZE, inp);
	}

	if (ferror (inp)) {
		fprintf (stderr, "%s: read error\n", argv0);
	}

	tmp = mdig;
	while (tmp != NULL) {
		n = tmp->get (tmp, buf, 256);

		if (par_bin) {
			if (fwrite (buf, 1, n, out) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}
		else {
			if (par_prefix) {
				fprintf (out, "%s:", tmp->name);
			}

			j = 0;
			for (i = 0; i < n; i++) {
				fputc (mdig_get_hex (buf[i] >> 4), out);
				j += 1;
				if ((par_digits > 0) && (j >= par_digits)) {
					fputc (' ', out);
					j = 0;
				}

				fputc (mdig_get_hex (buf[i] & 0x0f), out);
				j += 1;
				if ((par_digits > 0) && (j >= par_digits)) {
					fputc (' ', out);
					j = 0;
				}
			}

			fputc ('\n', out);
		}

		tmp = tmp->next;
	}

	fflush (out);

	return (0);
}

int main (int argc, char *argv[])
{
	int        i, r;
	const char *finp, *fout;
	FILE       *inp, *out;
	unsigned   fcnt;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help ();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("mdig");
			return (EXIT_SUCCESS);
		}
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing input file name\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing output file name\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "b", "binary")) {
			par_bin = 1;
		}
		else if (str_isarg2 (argv[i], "m", "digest")) {
			mdig_t *m;

			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing message digest\n", argv0
				);
				return (EXIT_FAILURE);
			}

			if (strcmp (argv[i], "size") == 0) {
				m = mdig_size_new();
			}
			else if (strcmp (argv[i], "sum8") == 0) {
				m = mdig_sum8_new();
			}
			else if (strcmp (argv[i], "adler8") == 0) {
				m = mdig_adler8_new();
			}
			else if (strcmp (argv[i], "adler16") == 0) {
				m = mdig_adler16_new();
			}
			else if (strcmp (argv[i], "adler32") == 0) {
				m = mdig_adler32_new();
			}
			else if (strcmp (argv[i], "crc") == 0) {
				m = mdig_crc_new();
			}
			else if (strcmp (argv[i], "md2") == 0) {
				m = mdig_md2_new();
			}
			else if (strcmp (argv[i], "md4") == 0) {
				m = mdig_md4_new();
			}
			else if (strcmp (argv[i], "md5") == 0) {
				m = mdig_md5_new();
			}
			else if (strcmp (argv[i], "rmd160") == 0) {
				m = mdig_rmd160_new();
			}
			else if (strcmp (argv[i], "sha1") == 0) {
				m = mdig_sha1_new();
			}
			else if (strcmp (argv[i], "sha224") == 0) {
				m = mdig_sha224_new();
			}
			else if (strcmp (argv[i], "sha256") == 0) {
				m = mdig_sha256_new();
			}
#ifndef MISC_DOS
			else if (strcmp (argv[i], "sha384") == 0) {
				m = mdig_sha384_new();
			}
			else if (strcmp (argv[i], "sha512") == 0) {
				m = mdig_sha512_new();
			}
#endif
			else if (strcmp (argv[i], "ed2k") == 0) {
				m = mdig_ed2k_new();
			}
			else {
				fprintf (stderr,
					"%s: unknown message digest (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}

			if (par_mdig_tl == NULL) {
				par_mdig = m;
			}
			else {
				par_mdig_tl->next = m;
			}

			par_mdig_tl = m;
		}
		else if (str_isarg2 (argv[i], "n", "digits")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing number of digits\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_digits)) {
				fprintf (stderr,
					"%s: bad number of digits (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "t", "prefix")) {
			par_prefix = 1;
		}
		else if (str_isarg2 (argv[i], "c", "capital")) {
			par_capital = 1;
		}
		else if (str_isarg2 (argv[i], "p", "pipe")) {
			par_pipe = 1;
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown argument (%s)\n",
				argv0, argv[i]
			);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	inp = (finp == NULL) ? stdin : fopen (finp, "rb");
	if (inp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n",
			argv0, finp
		);
		return (EXIT_FAILURE);
	}

	out = (fout == NULL) ? stdout : fopen (fout, "wb");
	if (out == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n",
			argv0, fout
		);
		return (EXIT_FAILURE);
	}

	r = mdig_calc (inp, out, par_mdig);

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
