/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/ngtotxt/ngtotxt.c                                        *
 * Created:     2003-09-28 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2003-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>

#include <stdlib.h>
#include <stdio.h>

#include "ngtotxt.h"


static const char *argv0 = NULL;


static int ng_entry (FILE *inp, FILE *out, unsigned long ofs, unsigned level);


static
void prt_help (void)
{
	fputs (
		"Usage: ngtotxt [options] [input [output]]\n"
		"  -i string          Set input file name [stdin]\n"
		"  -o string          Set output file name [stdout]\n",
		stdout
	);

	fflush (stdout);
}


static
unsigned short get_uint16 (const void *buf)
{
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	return (tmp[0] + (tmp[1] << 8));
}

static
unsigned long get_uint32 (const void *buf)
{
	unsigned long       ret;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	ret = tmp[3];
	ret = (ret << 8) | tmp[2];
	ret = (ret << 8) | tmp[1];
	ret = (ret << 8) | tmp[0];

	return (ret);
}

static
int ng_fgetc (FILE *fp)
{
	int c;

	c = fgetc (fp);
	if (c == EOF) {
		return (c);
	}

	return (c ^ 0x1a);
}

static
void ng_decrypt (void *buf, unsigned n)
{
	unsigned      i;
	unsigned char *tmp;

	tmp = (unsigned char *) buf;

	for (i = 0; i < n; i++) {
		tmp[i] ^= 0x1a;
	}
}

static
int ng_expand (unsigned char *dst, const unsigned char *src, unsigned max)
{
	unsigned i, n;

	i = 0;

	while (i < max) {
		switch (*src) {
			case 0x00:
				dst[i] = 0;
				return (0);

			case 0xff:
				src += 1;
				n = *src;

				while (n > 0) {
					dst[i++] = ' ';
					if (i >= max) {
						return (1);
					}
					n -= 1;
				}
				break;

			case '^':
				src += 1;
				if (*src == '^') {
					dst[i++] = '^';
				}
				break;

			default:
				dst[i++] = *src;
				break;
		}

		src += 1;
	}

	return (1);
}

static
int ng_get_uint16 (FILE *fp, unsigned short *ret)
{
	unsigned char buf[2];

	if (fread (buf, 1, 2, fp) != 2) {
		return (1);
	}

	ng_decrypt (buf, 2);

	*ret = get_uint16 (buf);

	return (0);
}

static
int ng_get_uint32 (FILE *fp, unsigned long *ret)
{
	unsigned char buf[4];

	if (fread (buf, 1, 4, fp) != 4) {
		return (1);
	}

	ng_decrypt (buf, 4);

	*ret = get_uint32 (buf);

	return (0);
}

static
int ng_get_string (unsigned char *dst, unsigned maxd, const void *src, unsigned maxs)
{
	unsigned      i;
	const char    *tmp;
	unsigned char buf[256];

	tmp = src;

	i = 0;
	while ((i < maxs) && (i < 256) && (tmp[i] != 0)) {
		buf[i] = tmp[i];
		i += 1;
	}

	if (i >= 256) {
		return (1);
	}

	buf[i] = 0;

	return (ng_expand (dst, buf, maxd));
}

static
int ng_get_line (FILE *fp, unsigned char *dst, unsigned max)
{
	unsigned      i;
	unsigned char buf[256];

	i = 0;

	do {
		buf[i++] = ng_fgetc (fp);
	} while ((i < 256) && (buf[i - 1] != 0));

	if (i >= 256) {
		return (1);
	}

	return (ng_expand (dst, buf, max));
}

static
void ng_print_ul (FILE *fp, const unsigned char *str, unsigned level)
{
	unsigned      i;
	unsigned char ul;

	while (*str == ' ') {
		str += 1;
	}

	fprintf (fp, "%s\n", str);

	switch (level) {
		case 0:
		case 1:
		case 2:
			ul = '+';
			break;

		default:
			ul = '-';
			break;
	}

	for (i = 0; i < 78; i++) {
		fputc (ul, fp);
	}

#if 0
	while (*str != 0) {
		fputc (ul, fp);
		str += 1;
	}
#endif

	fputs ("\n", fp);
}

static
int ng_header (FILE *inp, FILE *out)
{
	unsigned      i;
	unsigned char buf[378];
	unsigned char str[256];

	if (fread (buf, 1, 378, inp) != 378) {
		return (1);
	}

	if (get_uint16 (buf) != 0x474e) {
		return (1);
	}

	if (ng_get_string (str, 256, buf + 8, 40)) {
		return (1);
	}

	ng_print_ul (out, str, 0);
	fputs ("\n\n", out);

	for (i = 0; i < 5; i++) {
		if (ng_get_string (str, 256, buf + 48 + 66 * i, 66)) {
			return (1);
		}

		fprintf (out, "%s\n", str);
	}

	fputs ("\n", out);

	return (0);
}

static
int ng_short (FILE *inp, FILE *out, unsigned long ofs, unsigned level)
{
	unsigned       i;
	unsigned short type, size, lines, pline, tmp;
	unsigned long  ofs1, ofs2, entry;
	unsigned long  parent;
	unsigned char  buf[256];

	if (fseek (inp, ofs, SEEK_SET)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &type)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &size)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &lines)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &tmp)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &pline)) {
		return (1);
	}

	if (ng_get_uint32 (inp, &parent)) {
		return (1);
	}

	if (fseek (inp, ofs + 26 + 6 * lines, SEEK_SET)) {
		return (1);
	}

	ofs1 = ofs + 26;
	ofs2 = ofs + 26 + 6 * lines;

	for (i = 0; i < lines; i++) {
		if (fseek (inp, ofs2, SEEK_SET)) {
			return (1);
		}

		if (ng_get_line (inp, buf, 256)) {
			return (1);
		}

		ofs2 = ftell (inp);

		fputs ("\n\n", out);
		ng_print_ul (out, buf, level);
		fputs ("\n", out);

		if (fseek (inp, ofs1 + 2, SEEK_SET)) {
			return (1);
		}

		if (ng_get_uint32 (inp, &entry)) {
			return (1);
		}

		if (entry != 0xffffffffUL) {
			if (ng_entry (inp, out, entry, level + 1)) {
				return (1);
			}
		}

		ofs1 += 6;
	}

	return (0);
}

static
int ng_long (FILE *inp, FILE *out, unsigned long ofs, unsigned level)
{
	unsigned       i;
	unsigned short type, size, lines, see, parent;
	unsigned long  prev, next, tmp;
	unsigned char  buf[256];

	if (fseek (inp, ofs, SEEK_SET)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &type)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &size)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &lines)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &see)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &parent)) {
		return (1);
	}

	if (ng_get_uint32 (inp, &tmp)) {
		return (1);
	}

	if (ng_get_uint32 (inp, &tmp)) {
		return (1);
	}

	if (ng_get_uint32 (inp, &prev)) {
		return (1);
	}

	if (ng_get_uint32 (inp, &next)) {
		return (1);
	}

	for (i = 0; i < lines; i++) {
		if (ng_get_line (inp, buf, 256)) {
			return (1);
		}

		fprintf (out, "%s\n", buf);
	}

	if (see > 0) {
		if (ng_get_uint16 (inp, &see)) {
			return (1);
		}

		if (fseek (inp, 4 * see, SEEK_CUR)) {
			return (1);
		}

		fputs ("\nSee also: ", out);

		for (i = 0; i < see; i++) {
			if (ng_get_line (inp, buf, 256)) {
				return (1);
			}

			if (i > 0) {
				fputs (", ", out);
			}

			fprintf (out, "\"%s\"", buf);
		}

		fputs ("\n", out);
	}

	return (0);
}

static
int ng_entry (FILE *inp, FILE *out, unsigned long ofs, unsigned level)
{
	unsigned short type, size;

	if (fseek (inp, ofs, SEEK_SET)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &type)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &size)) {
		return (1);
	}

	if (type == 0) {
		return (ng_short (inp, out, ofs, level));
	}
	else if (type == 1) {
		return (ng_long (inp, out, ofs, level));
	}

	return (1);
}

static
int ng_menu (FILE *inp, FILE *out, unsigned long ofs)
{
	unsigned       i;
	unsigned short type, size, lines;
	unsigned char  str[256], title[256];
	unsigned long  ofs1, ofs2, entry;

	if (fseek (inp, ofs, SEEK_SET)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &type)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &size)) {
		return (1);
	}

	if (ng_get_uint16 (inp, &lines)) {
		return (1);
	}

	if (fseek (inp, ofs + 26 + 4 * (lines - 1) + 8 * lines, SEEK_SET)) {
		return (1);
	}

	if (ng_get_line (inp, title, 256)) {
		return (1);
	}

	fputs ("\n\n", out);
	ng_print_ul (out, title, 1);
	fputs ("\n", out);

	ofs1 = ofs + 26;
	ofs2 = ftell (inp);

	for (i = 1; i < lines; i++) {
		if (fseek (inp, ofs2, SEEK_SET)) {
			return (1);
		}

		if (ng_get_line (inp, str, 256)) {
			return (1);
		}

		ofs2 = ftell (inp);

		fputs ("\n\n", out);
		ng_print_ul (out, str, 2);
		fputs ("\n", out);

		if (fseek (inp, ofs1, SEEK_SET)) {
			return (1);
		}

		if (ng_get_uint32 (inp, &entry)) {
			return (1);
		}

		ofs1 += 4;

		if (ng_entry (inp, out, entry, 3)) {
			return (1);
		}
	}

	return (0);
}

static
int ng_body (FILE *inp, FILE *out, unsigned long ofs)
{
	unsigned short type, size;

	while (1) {
		if (fseek (inp, ofs, SEEK_SET)) {
			return (1);
		}

		if (ng_get_uint16 (inp, &type)) {
			return (0);
		}

		if (ng_get_uint16 (inp, &size)) {
			return (1);
		}

		if (type == 2) {
			if (ng_menu (inp, out, ofs)) {
				return (1);
			}
		}

		ofs += 26 + size;
	}

	return (0);
}

static
int ngtotxt (FILE *inp, FILE *out)
{
	if (ng_header (inp, out)) {
		return (1);
	}

	if (ng_body (inp, out, 378)) {
		return (1);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int      i, r;
	unsigned file_cnt;
	char     *fn_inp;
	char     *fn_out;
	FILE     *f_inp;
	FILE     *f_out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "-?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("ngtotxt");
			return (EXIT_SUCCESS);
		}
	}

	fn_inp = NULL;
	fn_out = NULL;

	file_cnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fn_inp = argv[i];
			file_cnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fn_out = argv[i];
			file_cnt = 2;
		}
		else if (argv[i][0] != '-') {
			if (file_cnt == 0) {
				fn_inp = argv[i];
				file_cnt = 1;
			}
			else if (file_cnt == 1) {
				fn_out = argv[i];
				file_cnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	f_inp = (fn_inp == NULL) ? stdin : fopen (fn_inp, "rb");
	if (f_inp == NULL) {
		fprintf (stderr, "%s: can't open input file\n", argv0);
		return (1);
	}

	f_out = (fn_out == NULL) ? stdout : fopen (fn_out, "wb");
	if (f_out == NULL) {
		fprintf (stderr, "%s: can't open output file\n", argv0);
		return (1);
	}

	r = ngtotxt (f_inp, f_out);

	if (f_out != stdout) {
		fclose (f_out);
	}

	if (f_inp != stdin) {
		fclose (f_inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
