/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/pbuf/pbuf.c                                              *
 * Created:     2005-12-15 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2005-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <errno.h>
#include <unistd.h>

#include "pbuf.h"


typedef struct pbuf_s {
	struct pbuf_s *next;
	unsigned      cnt;
	unsigned char buf[16384];
} pbuf_t;


static const char    *argv0 = NULL;

static int           par_pipe = 0;
static int           par_append = 0;
static unsigned long par_max_size = 0;


static
void prt_help (void)
{
	fputs (
		"pbuf: pipe buffer\n\n"
		"usage: pbuf [options] [input [output]]\n"
		"  -a, --append        Append to output file [no]\n"
		"  -i string           Set input file name [stdin]\n"
		"  -m, --max-size int  Set the maximum buffer size [0]\n"
		"  -o string           Set output file name [stdout]\n"
		"  -p, --pipe          Pass input through to stdout [no]\n",
		stdout
	);

	fflush (stdout);
}

static
pbuf_t *pbuf_new (void)
{
	pbuf_t *buf;

	buf = malloc (sizeof (pbuf_t));
	if (buf == NULL) {
		return (NULL);
	}

	buf->next = NULL;
	buf->cnt = 0;

	return (buf);
}

static
void pbuf_del (pbuf_t *buf)
{
	pbuf_t *tmp;

	while (buf != NULL) {
		tmp = buf;
		buf = buf->next;
		free (tmp);
	}
}

static
pbuf_t *pbuf_read_buf (FILE *inp, unsigned max)
{
	unsigned i;
	ssize_t  r;
	pbuf_t   *buf;

	buf = pbuf_new();
	if (buf == NULL) {
		fprintf (stderr, "%s: memory allocation error\n", argv0);
		return (NULL);
	}

	i = 0;
	while (i < max) {
		r = read (fileno (inp), buf->buf + i, max - i);

		if (r < 0) {
			if (errno != EINTR) {
				fprintf (stderr, "%s: read error\n", argv0);
				break;
			}
		}
		else if (r == 0) {
			break;
		}
		else {
			if (par_pipe) {
				fwrite (buf->buf + i, 1, r, stdout);
				fflush (stdout);
			}

			i += r;
		}
	}

	buf->cnt = i;

	if (i == 0) {
		pbuf_del (buf);
		return (NULL);
	}

	return (buf);
}

static
void pbuf_skip (FILE *inp)
{
	pbuf_t *buf;

	while (1) {
		buf = pbuf_read_buf (inp, sizeof (buf->buf));

		if (buf == NULL) {
			return;
		}

		pbuf_del (buf);
	}
}

static
pbuf_t *pbuf_read (FILE *inp)
{
	unsigned      n;
	unsigned long total;
	pbuf_t        *head, *tail, *buf;

	head = NULL;
	tail = NULL;

	total = 0;

	while (1) {
		n = sizeof (buf->buf);

		if (par_max_size > 0) {
			unsigned long m;

			if (total >= par_max_size) {
				pbuf_skip (inp);
				break;
			}

			m = (par_max_size - total);
			if (m < n) {
				n = m;
			}
		}

		/* n > 0 */

		buf = pbuf_read_buf (inp, n);
		if (buf == NULL) {
			break;
		}

		total += buf->cnt;

		if (head == NULL) {
			head = buf;
		}
		else {
			tail->next = buf;
		}

		tail = buf;
	}

	if (head == NULL) {
		head = pbuf_new();
	}

	return (head);
}

static
int pbuf_write (FILE *out, pbuf_t *buf)
{
	pbuf_t *tmp;

	while (buf != NULL) {
		tmp = buf;
		buf = buf->next;

		if (fwrite (tmp->buf, 1, tmp->cnt, out) != tmp->cnt) {
			fprintf (stderr, "%s: write error\n", argv0);
			pbuf_del (tmp);
			return (1);
		}

		free (tmp);
	}

	return (0);
}

static
int pbuf (const char *finp, const char *fout)
{
	FILE   *fp;
	pbuf_t *buf;

	if ((finp == NULL) || (strcmp (finp, "-") == 0)) {
		fp = stdin;
	}
	else {
		fp = fopen (finp, "rb");
		if (fp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n",
				argv0, finp
			);
			return (1);
		}
	}

	buf = pbuf_read (fp);
	if (buf == NULL) {
		return (1);
	}

	if (fp != stdin) {
		fclose (fp);
	}

	if ((fout == NULL) || (strcmp (fout, "-") == 0)) {
		fp = stdout;
	}
	else {
		fp = fopen (fout, par_append ? "ab" : "wb");
		if (fp == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n",
				argv0, fout
			);
			pbuf_del (buf);
			return (1);
		}
	}

	if (pbuf_write (fp, buf)) {
		return (1);
	}

	if (fp != stdout) {
		fclose (fp);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int        i;
	unsigned   fcnt;
	const char *finp, *fout;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("pbuf");
			return (EXIT_SUCCESS);
		}
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing input file name\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing output file name\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "a", "append")) {
			par_append = 1;
		}
		else if (str_isarg2 (argv[i], "m", "max-size")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing maximum size\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_max_size)) {
				fprintf (stderr, "%s: bad maximum size (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "p", "pipe")) {
			par_pipe = 1;
		}
		else if ((argv[i][0] != '-') || (argv[i][1] == 0)) {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n",
				argv0, argv[i]
			);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (pbuf (finp, fout)) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
