/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/portfwd/portfwd.c                                        *
 * Created:     2003-01-17 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2003-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/socket.h>
#include <lib/strarg.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/poll.h>

#include "portfwd.h"


static const char     *argv0 = NULL;

static int            par_term = 0;
static int            par_detach = 0;
static int            par_verbose = 0;

static unsigned       par_fwd_cnt = 0;
static portfwd_t      par_fwd[FWD_MAX_PORTS];
static struct pollfd  par_pfd[FWD_MAX_PORTS];

static char           *par_log = NULL;

static unsigned long  par_cct_cnt = 0;


static
void prt_help (void)
{
	fputs (
		"usage: portfwd [options]\n"
		"  --help                    Print usage information\n"
		"  --version                 Print version information\n"
		"  -v, --verbose             Verbose operation\n"
		"  -d, --detach              Detach\n"
		"  -t, --tcp port host:port  Forward a tcp port\n"
		"  -l, --log string          Set log file name (use '#')\n",
		stdout
	);

	fflush (stdout);
}

static
void sig_term (int sig)
{
	par_term = 1;
}

static
void sig_pipe (int sig)
{
}

static
void fwd_detach (void)
{
	int fd, msk;

	fd = open ("/dev/null", O_RDWR);

	msk = 0;

	close (0);
	close (1);

	if (isatty (2)) {
		close (2);
		msk |= 4;
	}

	if (fd >= 0) {
		dup2 (fd, 0);
		dup2 (fd, 1);

		if (msk & 4) {
			dup2 (fd, 2);
		}

		close (fd);
	}

	setpgid (0, 0);
}

static
void fwd_detach_and_fork (void)
{
	pid_t pid;

	fwd_detach();

	pid = fork();

	if (pid < 0) {
		exit (1);
	}
	else if (pid > 0) {
		exit (0);
	}
}

static
int fwd_resolve (const char *host, char **addr, unsigned short *port)
{
	char *tmp;

	if (tcp_addr_get_port (host, port)) {
		return (1);
	}

	tmp = tcp_addr_get_host (host);
	if (tmp == NULL) {
		return (1);
	}

	*addr = tcp_host_resolve (tmp);
	free (tmp);
	if (*addr == NULL) {
		return (1);
	}

	return (0);
}

static
int fwd_set_port (unsigned short port, const char *host)
{
	unsigned       i;
	char           *out_addr;
	unsigned short out_port;

	if (par_fwd_cnt >= FWD_MAX_PORTS) {
		return (1);
	}

	if (fwd_resolve (host, &out_addr, &out_port)) {
		return (1);
	}

	for (i = 0; i < par_fwd_cnt; i++) {
		if (par_fwd[i].inp_port == port) {
			free (par_fwd[i].out_addr);
			par_fwd[i].out_addr = out_addr;
			par_fwd[i].out_port = out_port;

			return (0);
		}
	}

	i = par_fwd_cnt;
	par_fwd[i].inp_addr = NULL;
	par_fwd[i].inp_port = port;
	par_fwd[i].inp_fd = -1;
	par_fwd[i].out_addr = out_addr;
	par_fwd[i].out_port = out_port;

	par_fwd_cnt += 1;

	return (0);
}

static
FILE *fwd_get_log (unsigned long cct)
{
	unsigned long i;
	char          *log;
	FILE          *ret;

	if (par_log == NULL) {
		return (NULL);
	}

	log = strdup (par_log);

	i = 0;
	while (log[i] != 0) {
		i += 1;
	}

	while (i > 0) {
		i -= 1;

		if (log[i] == '#') {
			log[i] = '0' + cct % 10;
			cct = cct / 10;
		}
	}

	ret = fopen (log, "wb");

	free (log);

	return (ret);
}

static
void fwd_log_connection (FILE *fp, int inp_fd, int out_fd)
{
	time_t         t;
	struct tm      *tm;
	char           *host[4];
	unsigned short port[4];

	host[0] = NULL;
	host[1] = NULL;
	host[2] = NULL;
	host[3] = NULL;

	tcp_get_peer (inp_fd, &host[0], &port[0]);
	tcp_get_host (inp_fd, &host[1], &port[1]);
	tcp_get_host (out_fd, &host[2], &port[2]);
	tcp_get_peer (out_fd, &host[3], &port[3]);

	t = time (NULL);
	tm = gmtime (&t);

	fprintf (fp,
		"%04d-%02d-%02d %02d:%02d:%02d  %s:%d -> %s:%d  %s:%d -> %s:%d\n",
		1900 + tm->tm_year, tm->tm_mon + 1, tm->tm_mday,
		tm->tm_hour, tm->tm_min, tm->tm_sec,
		(host[0] != NULL) ? host[0] : "<unknown>", port[0],
		(host[1] != NULL) ? host[1] : "<unknown>", port[1],
		(host[2] != NULL) ? host[2] : "<unknown>", port[2],
		(host[3] != NULL) ? host[3] : "<unknown>", port[3]
	);

	fflush (fp);

	free (host[3]);
	free (host[2]);
	free (host[1]);
	free (host[0]);
}

static
void fwd_log_data_line (FILE *fp, const unsigned char *buf, unsigned n, const char *prefix)
{
	unsigned i, j;

	if (n < 1) {
		return;
	}

	fputs (prefix, fp);
	fprintf (fp, "%02X", buf[0]);

	for (i = 1; i < n; i++) {
		fprintf (fp, " %02X", buf[i]);
	}

	j = strlen (prefix) + 3 * n - 1;

	do {
		fputs ("\t", fp);
		j = (j + 8) & ~7;
	} while (j < 56);

	fputs ("# ", fp);

	for (i = 0; i < n; i++) {
		if ((buf[i] >= 32) && (buf[i] < 128)) {
			fputc (buf[i], fp);
		}
		else {
			fputc ('.', fp);
		}
	}

	fputs ("\n", fp);
}

static
void fwd_log_data (FILE *fp, const unsigned char *buf, unsigned n, const char *prefix)
{
	unsigned i, j;

	i = 0;
	while (i < n) {
		j = n - i;
		j = (j < 16) ? j : 16;

		fwd_log_data_line (fp, buf + i, j, prefix);

		i += j;
	}

	fflush (fp);
}

static
int fwd_copy (int inp_fd, int out_fd, FILE *log, char *prefix)
{
	int           r;
	unsigned      i, n;
	unsigned char buf[FWD_BUF_SIZE];

	r = read (inp_fd, buf, FWD_BUF_SIZE);
	if (r <= 0) {
		return (1);
	}

	n = (unsigned) r;

	if (log != NULL) {
		fwd_log_data (log, buf, n, prefix);
	}

	i = 0;

	while (n > 0) {
		r = write (out_fd, buf + i, n);
		if (r <= 0) {
			return (1);
		}

		i += (unsigned) r;
		n -= (unsigned) r;
	}

	return (0);
}

static
int fwd_handle (int inp_fd, int out_fd, unsigned long cct)
{
	int           r;
	struct pollfd pfd[2];
	FILE          *log;

	log = fwd_get_log (cct);

	if (log != NULL) {
		fwd_log_connection (log, inp_fd, out_fd);
	}

	r = 1;

	while (par_term == 0) {
		pfd[0].fd = inp_fd;
		pfd[0].events = POLLIN;
		pfd[0].revents = 0;

		pfd[1].fd = out_fd;
		pfd[1].events = POLLIN;
		pfd[1].revents = 0;

		r = poll (pfd, 2, -1);
		if (r < 0) {
			break;
		}

		if (pfd[0].revents & POLLIN) {
			r = fwd_copy (inp_fd, out_fd, log, "<- ");
			if (r) {
				break;
			}
		}
		else if (pfd[1].revents & POLLIN) {
			r = fwd_copy (out_fd, inp_fd, log, "-> ");
			if (r) {
				break;
			}
		}
		else {
			break;
		}
	}

	if (log != NULL) {
		fclose (log);
	}

	return (r);
}

static
int fwd_accept (int inp_fd, const char *out_addr, unsigned short out_port)
{
	unsigned      i;
	int           r;
	int           out_fd;
	unsigned long cct;
	pid_t         pid;

	inp_fd = tcp_accept (inp_fd);
	if (inp_fd < 0) {
		return (1);
	}

	out_fd = tcp_connect (out_addr, out_port);
	if (out_fd < 0) {
		if (par_verbose) {
			fprintf (stderr, "%s: failed to connect to remote host (%s:%d)\n",
				argv0, out_addr, out_port
			);
		}
		close (inp_fd);
		return (0);
	}

	cct = par_cct_cnt++;

	if (par_verbose) {
		fwd_log_connection (stderr, inp_fd, out_fd);
	}

	pid = fork();

	if (pid < 0) {
		close (inp_fd);
		close (out_fd);
		return (1);
	}
	else if (pid == 0) {
		if (!par_detach) {
			fwd_detach();
		}

		for (i = 0; i < par_fwd_cnt; i++) {
			if (par_fwd[i].inp_fd >= 0) {
				close (par_fwd[i].inp_fd);
			}
		}

		r = fwd_handle (inp_fd, out_fd, cct);

		close (inp_fd);
		close (out_fd);

		exit (r);
	}
	else {
		close (inp_fd);
		close (out_fd);
		return (0);
	}

	return (1);
}

static
int fwd_poll (int tm)
{
	unsigned i, j, n;
	int      r;

	n = 0;
	for (i = 0; i < par_fwd_cnt; i++) {
		if (par_fwd[i].inp_fd >= 0) {
			par_pfd[n].fd = par_fwd[i].inp_fd;
			par_pfd[n].events = POLLIN;
			par_pfd[n].revents = 0;
			n += 1;
		}
	}

	r = poll (par_pfd, n, tm);
	if (r < 0) {
		return (1);
	}

	if (r == 0) {
		return (0);
	}

	for (i = 0; i < n; i++) {
		if (par_pfd[i].revents & POLLIN) {
			j = 0;
			while ((j < par_fwd_cnt) && (par_fwd[j].inp_fd != par_pfd[i].fd)) {
				j += 1;
			}

			if (j < par_fwd_cnt) {
				fwd_accept (par_fwd[j].inp_fd, par_fwd[j].out_addr, par_fwd[j].out_port);
			}
		}
	}

	return (0);
}

static
int fwd_listen (void)
{
	unsigned       i;
	int            fd;
	char           *addr;
	unsigned short port;

	for (i = 0; i < par_fwd_cnt; i++) {
		if (par_fwd[i].inp_fd < 0) {
			addr = par_fwd[i].inp_addr;
			port = par_fwd[i].inp_port;

			fd = tcp_listen (addr, port, 16);
			if (fd < 0) {
				fd = -1;
			}

			par_fwd[i].inp_fd = fd;
		}
	}

	return (0);
}

static
int fwd_wait (void)
{
	int   status;
	pid_t pid;

	pid = waitpid (-1, &status, WNOHANG);

	return (0);
}

int main (int argc, char *argv[])
{
	int            i;
	unsigned       j;
	unsigned short port;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help ();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("portfwd");
			return (EXIT_SUCCESS);
		}
	}

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "t", "tcp")) {
			if ((i + 3) > argc) {
				fprintf (stderr, "%s: missing host/port\n", argv0);
				return (EXIT_FAILURE);
			}

			port = (unsigned short) strtoul (argv[i + 1], NULL, 0);

			if (fwd_set_port (port, argv[i + 2])) {
				fprintf (stderr, "%s: bad host/port (%d -> %s)\n",
					argv0, port, argv[i + 2]
				);
				return (EXIT_FAILURE);
			}

			i += 2;
		}
		else if (str_isarg2 (argv[i], "l", "log")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing log file name\n", argv0);
				return (EXIT_FAILURE);
			}

			par_log = strdup (argv[i]);
		}
		else if (str_isarg2 (argv[i], "d", "detach")) {
			par_detach = 1;
		}
		else if (str_isarg2 (argv[i], "v", "verbose")) {
			par_verbose = 1;
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	signal (SIGINT, &sig_term);
	signal (SIGTERM, &sig_term);
	signal (SIGPIPE, &sig_pipe);

	if ((par_fwd_cnt > 0) && (par_term == 0)) {
		if (par_detach) {
			fwd_detach_and_fork();
		}

		while (par_term == 0) {
			fwd_listen();
			fwd_poll (1000);
			fwd_wait();
		}
	}

	for (j = 0; j < par_fwd_cnt; j++) {
		if (par_fwd[j].inp_fd >= 0) {
			close (par_fwd[j].inp_fd);
		}
	}

	if (par_verbose) {
		fprintf (stderr, "%s: terminating\n", argv0);
	}

	return (EXIT_SUCCESS);
}
