/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/quote/quote.c                                            *
 * Created:     2004-05-21 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2004-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include "quote.h"


static const char    *argv0 = NULL;

static unsigned      par_enc = 1;
static unsigned      par_quote = '=';
static unsigned char par_chars[256];


static
void prt_help (void)
{
	fputs (
		"Usage: quote [options] [input [output]]\n"
		"  -a, --quote-char char           Quote a character\n"
		"  -A, --quote-range char char     Quote a range of characters\n"
		"  -d, --decode                    Unquote\n"
		"  -e, --encode                    Quote [default]\n"
		"  -i string                       Set input file name [stdin]\n"
		"  -o string                       Set output file name [stdout]\n"
		"  -p, --quoted-printable          Set defaults for quoted-printable [no]\n"
		"  -q, --quote char                Set the quote character [=]\n"
		"  -r, --no-quote-char char        Don't quote a character\n"
		"  -R, --no-quote-range char char  Don't quote a range of characters\n",
		stdout
	);

	fflush (stdout);
}

static
void put_hex (FILE *fp, unsigned val)
{
	if (val < 10) {
		fputc ('0' + val, fp);
	}
	else {
		fputc ('a' + (val - 10), fp);
	}
}

static
int get_hex (FILE *fp, unsigned *val)
{
	unsigned i;
	int      c;

	*val = 0;

	for (i = 0; i < 2; i++) {
		c = fgetc (fp);
		if ((c >= '0') && (c <= '9')) {
			*val = (*val << 4) + (c - '0');
		}
		else if ((c >= 'a') && (c <= 'f')) {
			*val = (*val << 4) + (c - 'a' + 10);
		}
		else if ((c >= 'A') && (c <= 'F')) {
			*val = (*val << 4) + (c - 'A' + 10);
		}
		else {
			return (1);
		}
	}

	return (0);
}

static
int get_char (const char *str, unsigned *val)
{
	if ((str[0] >= '0') && (str[0] <= '9')) {
		if (str_get_uint (str, val)) {
			return (1);
		}
	}
	else {
		*val = str[0];
	}

	return (0);
}

static
void set_default_quote (void)
{
	unsigned i;

	for (i = 0; i < 32; i++) {
		par_chars[i] = 1;
	}

	for (i = 32; i < 128; i++) {
		par_chars[i] = 0;
	}

	for (i = 128; i < 256; i++) {
		par_chars[i] = 1;
	}

	par_chars[0x09] = 0;
	par_chars[0x0a] = 0;
	par_chars[0x0d] = 0;

	par_quote = '=';
}

static
void set_quoted_printable (void)
{
	unsigned i;

	for (i = 0; i < 32; i++) {
		par_chars[i] = 1;
	}

	for (i = 32; i < 128; i++) {
		par_chars[i] = 0;
	}

	for (i = 128; i < 256; i++) {
		par_chars[i] = 1;
	}

	par_chars[0x0a] = 0;
	par_chars[0x0d] = 0;
	par_chars[0x20] = 1;

	par_quote = '=';
}

static
int quote_enc (FILE *inp, FILE *out, unsigned char q)
{
	int c;

	c = fgetc (inp);
	while (c != EOF) {
		if ((par_chars[c] != 0) || (c == q)) {
			fputc (q, out);
			put_hex (out, (c >> 4) & 0x0f);
			put_hex (out, c & 0x0f);
		}
		else {
			fputc (c, out);
		}

		c = fgetc (inp);
	}

	return (0);
}

static
int quote_dec (FILE *inp, FILE *out, unsigned char q)
{
	int      c;
	unsigned val;

	c = fgetc (inp);
	while (c != EOF) {
		if (c == q) {
			if (get_hex (inp, &val)) {
				return (1);
			}

			fputc (val, out);
		}
		else {
			fputc (c, out);
		}

		c = fgetc (inp);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int        i, r;
	unsigned   fcnt;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("quote");
			return (EXIT_SUCCESS);
		}
	}

	set_default_quote();

	fcnt = 0;
	finp = NULL;
	fout = NULL;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "e", "encode")) {
			par_enc = 1;
		}
		else if (str_isarg2 (argv[i], "d", "decode")) {
			par_enc = 0;
		}
		else if (str_isarg2 (argv[i], "q", "quote")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing quote character\n", argv0);
				return (EXIT_FAILURE);
			}

			if (get_char (argv[i], &par_quote)) {
				fprintf (stderr, "%s: bad quote character (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "a", "quote-char")) {
			unsigned c;

			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing character\n", argv0);
				return (EXIT_FAILURE);
			}

			if (get_char (argv[i], &c)) {
				fprintf (stderr, "%s: bad character (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			par_chars[c & 0xff] = 1;
		}
		else if (str_isarg2 (argv[i], "A", "quote-range")) {
			unsigned c1, c2;

			if ((i + 2) >= argc) {
				fprintf (stderr, "%s: missing character range\n", argv0);
				return (EXIT_FAILURE);
			}

			if (get_char (argv[i + 1], &c1) || get_char (argv[i + 2], &c2)) {
				fprintf (stderr, "%s: bad character range (%s - %s)\n",
					argv0, argv[i + 1], argv[i + 2]
				);
				return (EXIT_FAILURE);
			}

			c1 &= 0xff;
			c2 &= 0xff;

			while (c1 <= c2) {
				par_chars[c1++] = 1;
			}

			i += 2;
		}
		else if (str_isarg2 (argv[i], "p", "quoted-printable")) {
			set_quoted_printable();
		}
		else if (str_isarg2 (argv[i], "r", "no-quote-char")) {
			unsigned c;

			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing character\n", argv0);
				return (EXIT_FAILURE);
			}

			if (get_char (argv[i], &c)) {
				fprintf (stderr, "%s: bad character (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			par_chars[c & 0xff] = 0;
		}
		else if (str_isarg2 (argv[i], "R", "no-quote-range")) {
			unsigned c1, c2;

			if ((i + 2) >= argc) {
				fprintf (stderr, "%s: missing character range\n", argv0);
				return (EXIT_FAILURE);
			}

			if (get_char (argv[i + 1], &c1) || get_char (argv[i + 2], &c2)) {
				fprintf (stderr, "%s: bad character range (%s - %s)\n",
					argv0, argv[i + 1], argv[i + 2]
				);
				return (EXIT_FAILURE);
			}

			c1 &= 0xff;
			c2 &= 0xff;

			while (c1 <= c2) {
				par_chars[c1++] = 0;
			}

			i += 2;
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (finp == NULL) {
		inp = stdin;
	}
	else {
		inp = fopen (finp, "rb");
		if (inp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
			return (EXIT_FAILURE);
		}
	}

	if (fout == NULL) {
		out = stdout;
	}
	else {
		out = fopen (fout, "wb");
		if (out == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
			return (EXIT_FAILURE);
		}
	}

	if (par_enc) {
		r = quote_enc (inp, out, par_quote);
	}
	else {
		r = quote_dec (inp, out, par_quote);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (out != stdout) {
		fclose (out);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
