/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/readbin/readbin.c                                        *
 * Created:     2002-05-05 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <stdlib.h>

#include "readbin.h"


#define MISC_READBIN_BUFFER MISC_BUF_SIZE


static const char *argv0 = NULL;

static unsigned par_base = 10;
static int      par_little = 0;


static
void prt_help (void)
{
	fputs (
		"usage: readbin [options]\n"
		"  -i string            Set input file name [stdin]\n"
		"  -o string            Set output file name [stdout]\n"
		"  -n, --base int       Set number system base [10]\n"
		"  -p, --position int   Set absolute file position\n"
		"  -b, --big-endian     Use big endian [default]\n"
		"  -l, --little-endian  Use little endian\n"
		"  -u, --unsigned m n   Get next m * n bytes as m byte unsigned integers\n"
		"  -s, --signed m n     Get next m * n bytes as m byte signed integers\n"
		"  -k, --skip n         Skip n bytes\n"
		"  -r, --read n         Read n bytes\n"
		"  -z, --readz n        Read n bytes and fill with zero\n"
		"  -c, --cat            Read rest of input file\n"
		"  --uint8              Read a 8 bit unsigned integer\n"
		"  --uint16             Read a 16 bit unsigned integer\n"
		"  --uint32             Read a 32 bit unsigned integer\n"
		"  --sint8              Read a 8 bit signed integer\n"
		"  --sint16             Read a 16 bit signed integer\n"
		"  --sint32             Read a 32 bit signed integer\n",
		stdout
	);

	fflush (stdout);
}

static
void prt_uint32 (FILE *out, unsigned long val)
{
	unsigned i;
	unsigned dig;
	char     buf[256];

	i = 0;

	do {
		dig = (unsigned) (val % par_base);
		val = val / par_base;

		if (dig < 10 ) {
			buf[i] = (char) ('0' + dig);
		}
		else {
			buf[i] = (char) ('A' + dig - 10);
		}

		i += 1;
	} while ((val != 0) && (i < 256));

	while (i > 0) {
		i -= 1;
		fputc (buf[i], out);
	}

	fputc ('\n', out);

	fflush (out);
}

static
void prt_sint32 (FILE *out, long val)
{
	if (val < 0) {
		val = -val;
		fputc ('-', out);
	}

	prt_uint32 (out, (unsigned long) val);
}

static
unsigned get_uint8 (unsigned char c[1])
{
	return (c[0]);
}

static
unsigned get_uint16 (unsigned char c[2])
{
	unsigned r;

	if (par_little) {
		r = c[1] & 0xff;
		r = (r << 8) + (c[0] & 0xff);
	}
	else {
		r = c[0] & 0xff;
		r = (r << 8) + (c[1] & 0xff);
	}

	return (r);
}

static
unsigned long get_uint32 (unsigned char c[4])
{
	unsigned long r;

	if (par_little) {
		r = c[3] & 0xff;
		r = (r << 8) + (c[2] & 0xff);
		r = (r << 8) + (c[1] & 0xff);
		r = (r << 8) + (c[0] & 0xff);
	}
	else {
		r = c[0] & 0xff;
		r = (r << 8) + (c[1] & 0xff);
		r = (r << 8) + (c[2] & 0xff);
		r = (r << 8) + (c[3] & 0xff);
	}

	return (r);
}

static
int get_sint8 (unsigned char c[1])
{
	int r;

	r = c[0] & 0xff;

	if (r > 127) {
		r = r - 256;
	}

	return (r);
}

static
int get_sint16 (unsigned char c[2])
{
	long     r;

	r = (long) get_uint16 (c);

	if (r > 32767) {
		r = r - 65536;
	}

	return ((int) r);
}

static
long get_sint32 (unsigned char c[4])
{
	long          r;
	unsigned long tmp;

	tmp = get_uint32 (c);

	if (tmp >= (1 << 31)) {
		r = (long) (tmp - (1 << 31));
		r = r - (1 << 31);
	}
	else {
		r = (long) tmp;
	}

	return (r);
}

static
int rb_bytes (FILE *inp, FILE *out, unsigned long n, int fill)
{
	unsigned long i;
	unsigned char buf[MISC_READBIN_BUFFER];

	while (n > 0) {
		i = (n < MISC_READBIN_BUFFER) ? n : MISC_READBIN_BUFFER;

		i = fread (buf, 1, i, inp);
		if (i == 0) {
			break;
		}

		if (fwrite (buf, 1, i, out) != i) {
			return (1);
		}

		n -= i;
	}

	if (n == 0) {
		return (0);
	}

	if (fill == 0) {
		return (1);
	}

	memset (buf, 0, MISC_READBIN_BUFFER);

	while (n > 0) {
		i = (n < MISC_READBIN_BUFFER) ? n : MISC_READBIN_BUFFER;

		if (fwrite (buf, 1, i, out) != i) {
			return (1);
		}

		n -= i;
	}

	return (0);
}

static
int rb_skip (FILE *inp, FILE *out, unsigned long n)
{
	unsigned long i;
	unsigned char buf[MISC_READBIN_BUFFER];

	if (n <= 0x7fffffffUL) {
		if (fseek (inp, (long) n, SEEK_CUR) == 0) {
			return (0);
		}
	}

	while (n > 0) {
		i = (n < MISC_READBIN_BUFFER) ? n : MISC_READBIN_BUFFER;

		i = fread (buf, 1, i, inp);
		if (i == 0) {
			return (0);
		}

		n -= i;
	}

	return (0);
}

static
int rb_pos (FILE *inp, FILE *out, unsigned long n)
{
	if (n <= 0x7fffffffUL) {
		if (fseek (inp, (long) n, SEEK_SET) == 0) {
			return (0);
		}
	}

	return (1);
}

static
int rb_cat (FILE *inp, FILE *out)
{
	unsigned long n;
	unsigned char buf[MISC_READBIN_BUFFER];

	n = fread (buf, 1, MISC_READBIN_BUFFER, inp);

	while (n > 0) {
		if (fwrite (buf, 1, n, out) != n) {
			return (1);
		}

		n = fread (buf, 1, MISC_READBIN_BUFFER, inp);
	}

	return (0);
}

static
int rb_int8 (FILE *inp, FILE *out, int sign, unsigned long n)
{
	unsigned long i;
	unsigned char buf[1];

	for (i = 0; i < n; i++) {
		if (fread (buf, 1, 1, inp) != 1) {
			return (1);
		}

		if (sign) {
			prt_sint32 (out, get_sint8 (buf));
		}
		else {
			prt_uint32 (out, get_uint8 (buf));
		}
	}

	return (0);
}

static
int rb_int16 (FILE *inp, FILE *out, int sign, unsigned long n)
{
	unsigned long i;
	unsigned char buf[2];

	for (i = 0; i < n; i++) {
		if (fread (buf, 1, 2, inp) != 2) {
			return (1);
		}

		if (sign) {
			prt_sint32 (out, get_sint16 (buf));
		}
		else {
			prt_uint32 (out, get_uint16 (buf));
		}
	}

	return (0);
}

static
int rb_int32 (FILE *inp, FILE *out, int sign, unsigned long n)
{
	unsigned long i;
	unsigned char buf[4];

	for (i = 0; i < n; i++) {
		if (fread (buf, 1, 4, inp) != 4) {
			return (1);
		}

		if (sign) {
			prt_sint32 (out, get_sint32 (buf));
		}
		else {
			prt_uint32 (out, get_uint32 (buf));
		}
	}

	return (0);
}

static
int rb_unsigned (FILE *inp, FILE *out, unsigned long m, unsigned long n)
{
	switch (m) {
		case 0:
			return (1);

		case 1:
			return (rb_int8 (inp, out, 0, n));

		case 2:
			return (rb_int16 (inp, out, 0, n));

		case 4:
			return (rb_int32 (inp, out, 0, n));

		default:
			return (1);
	}

	return (1);
}

static
int rb_signed (FILE *inp, FILE *out, unsigned long m, unsigned long n)
{
	switch (m) {
		case 0:
			return (1);

		case 1:
			return (rb_int8 (inp, out, 1, n));

		case 2:
			return (rb_int16 (inp, out, 1, n));

		case 4:
			return (rb_int32 (inp, out, 1, n));

		default:
			return (1);
	}

	return (1);
}

int main (int argc, char *argv[])
{
	int           i;
	unsigned long m, n;
	FILE          *finp, *fout;


	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("readbin");
			return (EXIT_SUCCESS);
		}
	}

	finp = stdin;
	fout = stdout;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			if (finp != stdin) {
				fclose (finp);
			}

			finp = fopen (argv[i], "rb");

			if (finp == NULL) {
				fprintf (stderr, "%s: can't open input file (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			if ((fout != stdout) && (fout != stderr)) {
				fclose (fout);
			}

			fout = fopen (argv[i], "wb");

			if (fout == NULL) {
				fprintf (stderr, "%s: can't open output file (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "l", "little-endian")) {
			par_little = 1;
		}
		else if (str_isarg2 (argv[i], "b", "big-endian")) {
			par_little = 0;
		}
		else if (str_isarg2 (argv[i], "n", "base")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing base\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_base)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "k", "skip")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing skip count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &n)) {
				fprintf (stderr, "%s: illegal number (%s)\n", argv0, argv[i]);
				return (1);
			}

			if (rb_skip (finp, fout, n)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "p", "position")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing position\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &n)) {
				fprintf (stderr, "%s: illegal number (%s)\n", argv0, argv[i]);
				return (1);
			}

			if (rb_pos (finp, fout, n)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "r", "read")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing read count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &n)) {
				fprintf (stderr, "%s: illegal number (%s)\n", argv0, argv[i]);
				return (1);
			}

			if (rb_bytes (finp, fout, n, 0)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "z", "readz")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing read count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &n)) {
				fprintf (stderr, "%s: illegal number (%s)\n", argv0, argv[i]);
				return (1);
			}

			if (rb_bytes (finp, fout, n, 1)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "c", "cat")) {
			if (rb_cat (finp, fout)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "u", "unsigned")) {
			if ((i + 2) >= argc) {
				fprintf (stderr, "%s: missing unsigned count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i + 1], &m)) {
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i + 2], &n)) {
				return (EXIT_FAILURE);
			}

			if (rb_unsigned (finp, fout, m, n)) {
				return (EXIT_FAILURE);
			}

			i += 2;
		}
		else if (str_isarg2 (argv[i], "s", "signed")) {
			if ((i + 2) >= argc) {
				fprintf (stderr, "%s: missing signed count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i + 1], &m)) {
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i + 2], &n)) {
				return (EXIT_FAILURE);
			}

			if (rb_signed (finp, fout, m, n)) {
				return (EXIT_FAILURE);
			}

			i += 2;
		}
		else if (str_isarg2 (argv[i], NULL, "uint8")) {
			if (rb_int8 (finp, fout, 0, 1)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], NULL, "uint16")) {
			if (rb_int16 (finp, fout, 0, 1)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], NULL, "uint32")) {
			if (rb_int32 (finp, fout, 0, 1)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], NULL, "sint8")) {
			if (rb_int8 (finp, fout, 1, 1)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], NULL, "sint16")) {
			if (rb_int16 (finp, fout, 1, 1)) {
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], NULL, "sint32")) {
			if (rb_int32 (finp, fout, 1, 1)) {
				return (EXIT_FAILURE);
			}
		}
		else if (argv[i][0] != '-') {
			if (finp != stdin) {
				fclose (finp);
			}

			finp = fopen (argv[i], "rb");

			if (finp == NULL) {
				fprintf (stderr, "%s: can't open input file (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if ((fout != stdout) && (fout != stderr)) {
		fclose (fout);
	}

	if (finp != stdin) {
		fclose (finp);
	}

	return (EXIT_SUCCESS);
}
