/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/sleep/sleep.c                                            *
 * Created:     2001-12-11 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2001-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include "sleep.h"


static const char         *argv0 = NULL;

static unsigned long long par_time = 0;
static int                par_verbose = 0;


static
void prt_help (void)
{
	fputs (
		"usage: sleep [options] [time...]\n"
		"  -v, --verbose  Verbose operation\n"
		"Time is specified in seconds except if one of these suffixes is used:\n"
		"  d   days\n"
		"  h   hours\n"
		"  m   minutes\n"
		"  s   seconds\n"
		"  ms  milliseconds\n"
		"  u   microseconds\n",
		stdout
	);

	fflush (stdout);
}

static
int is_digit (char c)
{
	if ((c >= '0') && (c <= '9')) {
		return (1);
	}

	return (0);
}

static
int time_from_str (unsigned long long *tm, const char *str)
{
	unsigned           i;
	unsigned           dig;
	unsigned long long val;

	if (strchr (str, '.') != NULL) {
		double val;
		char   *end;

		val = strtod (str, &end);
		if ((end != NULL) && (*end != 0)) {
			return (1);
		}

		*tm = (unsigned long long) (1.0e6 * val);

		return (0);
	}

	val = 0;

	i = 0;
	while (is_digit (str[i])) {
		dig = (unsigned) (str[i] - '0');

		val = 10 * val + dig;

		i += 1;
	}

	if (str[i] == 0) {
		*tm = 1000000 * val;
		return (0);
	}

	switch (str[i]) {
		case 'd':
			val *= 24 * 60 * 60;
			val *= 1000000;
			break;

		case 'h':
			val *= 60 * 60;
			val *= 1000000;
			break;

		case 'm':
			if (str[i + 1] == 's') {
				i += 1;
				val *= 1000;
			}
			else {
				val *= 60;
				val *= 1000000;
			}
			break;

		case 'u':
			if (str[i + 1] == 's') {
				i += 1;
			}
			break;

		case 's':
			val *= 1000000;
			break;

		default:
			return (1);
	}

	*tm = val;

	if (str[i + 1] != 0) {
		return (1);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int                i;
	unsigned long long val;
	unsigned long      sleep_s;
	unsigned long      sleep_u;
	struct timeval     tv;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("sleep");
			return (EXIT_SUCCESS);
		}
	}

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "v", "verbose")) {
			par_verbose = 1;
		}
		else if (!is_digit (argv[i][0])) {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}
		else {
			if (time_from_str (&val, argv[i])) {
				fprintf (stderr, "%s: bad time format (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			par_time += val;
		}

		i += 1;
	}

	sleep_u = (unsigned long) (par_time % 1000000);
	sleep_s = (unsigned long) (par_time / 1000000);

	if (par_verbose) {
		fprintf (stdout, "sleeping %lu seconds + %lu microseconds\n", sleep_s, sleep_u);
	}

	while (sleep_s > 0) {
		sleep_s = sleep ((unsigned) sleep_s);
	}

	if (sleep_u > 0) {
		tv.tv_sec = 0;
		tv.tv_usec = (long) sleep_u;

		select (0, NULL, NULL, NULL, &tv);
	}

	return (EXIT_SUCCESS);
}
