/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/slice/slice.c                                            *
 * Created:     1999-03-25 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 1999-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "slice.h"


#define SLC_BUF_SIZE MISC_BUF_SIZE


static const char    *argv0 = NULL;

static lst_name_t    *par_lstn = NULL;
static lst_size_t    *par_lsts = NULL;
static char          *par_inp = NULL;
static unsigned long par_first = 1;
static int           par_verbose = 0;


static
void prt_help (void)
{
	fputs (
		"Usage: slice [options] [input [output...]]\n"
		"  -f, --first int  Set the number of the first file [1]\n"
		"  -i string        Set input file name [stdin]\n"
		"  -o string        Set output file name\n"
		"  -s, --size int   Set file size [1M]\n"
		"  -v, --verbose    Verbose operation [no]\n",
		stdout
	);

	fflush (stdout);
}

static
int lstn_add (lst_name_t **lst, const char *name)
{
	lst_name_t *el, *tmp;

	el = malloc (sizeof (lst_name_t));
	if (el == NULL) {
		return (1);
	}

	el->next = NULL;
	el->name = strdup (name);

	if (*lst == NULL) {
		*lst = el;
	}
	else {
		tmp = *lst;
		while (tmp->next != NULL) {
			tmp = tmp->next;
		}

		tmp->next = el;
	}

	return (0);
}

static
int lsts_add (lst_size_t **lst, unsigned long n)
{
	lst_size_t *el, *tmp;

	el = malloc (sizeof (lst_size_t));
	if (el == NULL) {
		return (1);
	}

	el->next = NULL;
	el->size = n;

	if (*lst == NULL) {
		*lst = el;
	}
	else {
		tmp = *lst;
		while (tmp->next != NULL) {
			tmp = tmp->next;
		}

		tmp->next = el;
	}

	return (0);
}

static
char *slice_get_name (const char *str, unsigned long n)
{
	unsigned long i;
	char          *ret;

	ret = strdup (str);

	i = strlen (ret);

	while (i > 0) {
		i -= 1;

		if (ret[i] == '#') {
			ret[i] = '0' + (n % 10);
			n = n / 10;
		}
	}

	return (ret);
}

static
int slice_file (FILE *inp, const char *fout, unsigned long size)
{
	FILE          *out;
	unsigned long i, k, n;
	size_t        r;
	unsigned char buf[SLC_BUF_SIZE];
	time_t        t1;

	out = NULL;

	i = 0;
	k = size;

	t1 = 0;

	while (size > 0) {
		n = (size < SLC_BUF_SIZE) ? size : SLC_BUF_SIZE;

		r = fread (buf, 1, n, inp);
		if (r <= 0) {
			if (out != NULL) {
				fclose (out);
			}

			if (ferror (inp)) {
				fprintf (stderr, "%s: read error\n", argv0);
				return (1);
			}

			if (par_verbose) {
				fprintf (stderr, "[%5.1f%%] %s (%lu / %lu)  \n",
					100.0, fout, i, i
				);
			}

			return (0);
		}

		if (out == NULL) {
			out = fopen (fout, "wb");
			if (out == NULL) {
				fprintf (stderr,
					"%s: can't create output file (%s)\n",
					argv0, fout
				);
				return (1);
			}
		}

		if (fwrite (buf, 1, r, out) != r) {
			fprintf (stderr, "%s: write error\n", argv0);
			fclose (out);
			return (1);
		}

		size -= r;

		i += r;

		if (par_verbose) {
			time_t t2;

			t2 = time (NULL);

			if (t1 != t2) {
				t1 = t2;
				fprintf (stderr, "[%5.1f%%] %s (%lu / %lu)\r",
					100.0 * (double) i / (double) k, fout, i, k
				);
			}
		}
	}

	if (out != NULL) {
		fclose (out);
	}

	if (par_verbose) {
		fprintf (stderr, "[%5.1f%%] %s (%lu / %lu)\n",
			100.0, fout, i, k
		);
	}
	return (0);
}

static
int slice (const char *in, lst_name_t *lstn, lst_size_t *lsts)
{
	unsigned long idx;
	lst_name_t    *curname;
	lst_size_t    *cursize;
	FILE          *inp;
	char          *fname;

	if (strcmp (in, "-") == 0) {
		inp = stdin;
	}
	else {
		inp = fopen (in, "rb");
		if (inp == NULL) {
			fprintf (stderr,
				"%s: can't open input file (%s)\n",
				argv0, in
			);
			return (1);
		}
	}

	idx = par_first;

	curname = lstn;
	cursize = lsts;

	while (!feof (inp)) {
		fname = slice_get_name (curname->name, idx);

		if (slice_file (inp, fname, cursize->size)) {
			if (inp != stdin) {
				fclose (inp);
			}
			free (fname);
			return (1);
		}

		free (fname);

		if (curname->next != NULL) {
			curname = curname->next;
		}

		if (cursize->next != NULL) {
			cursize = cursize->next;
		}

		idx += 1;
	}

	if (inp != stdin) {
		fclose (inp);
	}

	return (0);
}

int main (int argc, char **argv)
{
	int      i, r;
	unsigned file_cnt;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("slice");
			return (EXIT_SUCCESS);
		}
	}

	file_cnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing input file name\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			par_inp = strdup (argv[i]);

			file_cnt += 1;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing output file name\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			lstn_add (&par_lstn, argv[i]);

			file_cnt += 1;
		}
		else if (str_isarg2 (argv[i], "s", "size")) {
			unsigned long n;

			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing output file size\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &n)) {
				fprintf (stderr,
					"%s: illegal number (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}

			if (n > 0) {
				lsts_add (&par_lsts, n);
			}
		}
		else if (str_isarg2 (argv[i], "f", "first")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing first file number\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_first)) {
				fprintf (stderr,
					"%s: illegal number (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "v", "verbose")) {
			par_verbose = 1;
		}
		else if (argv[i][0] != '-') {
			if (file_cnt == 0) {
				par_inp = strdup (argv[i]);
			}
			else {
				lstn_add (&par_lstn, argv[i]);
			}

			file_cnt += 1;
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n",
				argv0, argv[i]
			);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (par_inp == NULL) {
		par_inp = strdup ("-");
	}

	if (par_lstn == NULL) {
		lstn_add (&par_lstn, "slice####.out");
	}

	if (par_lsts == NULL) {
		lsts_add (&par_lsts, 1024UL * 1024UL);
	}

	r = slice (par_inp, par_lstn, par_lsts);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
