/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/slist/slist.c                                            *
 * Created:     2007-01-12 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2007-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>
#include <lib/md5.h>

#include "slist.h"

#include <dirent.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>


static const char    *argv0 = NULL;

static int           par_files_first = 0;
static int           par_list_dir = 0;
static int           par_list_dir_link = 1;
static int           par_list_long = 0;
static int           par_list_md5 = 0;
static int           par_link_md5 = 0;
static int           par_list_inode = 0;
static int           par_list_tz = 0;
static int           par_flush = 0;
static int           par_reverse = 0;
static int           par_xdev = 0;
static dev_t         par_xdev_dev;
static int           par_utc = 1;
static int           par_noquote = 0;
static int           par_follow_symlink = 0;
static const char    *par_sort_files = NULL;
static const char    *par_sort_dirs = NULL;

static int                par_stats = 0;
static unsigned long      par_stat_entries = 0;
static unsigned long      par_stat_dirs = 0;
static unsigned long      par_stat_links = 0;
static unsigned long      par_stat_files = 0;
static unsigned long long par_stat_total = 0;
static unsigned long long par_stat_min = 0;
static unsigned long long par_stat_max = 0;

static char          *par_prefix = NULL;

static char          par_buf[16384];
static unsigned      par_buf_size = 16384;


typedef struct {
	char               *name;
	char               *qname;
	char               *link;
	char               *qlink;
	int                isdir;
	mode_t             mode;
	mode_t             lmode;
	time_t             mtime;
	unsigned long long size;
	dev_t              dev;
	ino_t              inode;
} slist_t;


static
void prt_help (void)
{
	fputs (
		"slist: sorted directory listing\n"
		"\n"
		"usage: slist [options] [dir...]\n"
		"  -d, --directories              List directory entries [no]\n"
		"  -D, --no-dir-symlinks          Don't list symlinks to directories\n"
		"  -f, --files-first              List files before directories [no]\n"
		"  -F, --flush                    Flush output after every file [no]\n"
		"  -i string                      List a directory\n"
		"  -I, --inodes                   List device and inode information [no]\n"
		"  -l, --long                     Long listing format [no]\n"
		"  -L, --symlink-md5              List MD5 for symlinks [no]\n"
		"  -m, --md5                      Include MD5 in listing [no]\n"
		"  -p, --prefix string            Add prefix to file names [none]\n"
		"  -q, --no-quote                 Don't quote file names [no]\n"
		"  -r, --reverse                  Reverse sort order [no]\n"
		"  -s, --sort-files string        Set sort key for files [+n]\n"
		"  -S, --sort-directories string  Set sort key for directories [+n]\n"
		"  -t, --stats                    Print statistics at the end [no]\n"
		"  -u, --no-utc                   List times in local time [no]\n"
		"  -x, --same-device              Stay on same device [no]\n"
		"  -y, --follow-symlinks          Follow symlinks [no]\n"
		"  -z, --timezone                 List the time zone [no]\n",
		stdout
	);

	fflush (stdout);
}

static
void slist_get_mode (slist_t *ent, char *str)
{
	if (ent->link != NULL) {
		str[0] = 'l';
	}
	else if (ent->isdir) {
		str[0] = 'd';
	}
	else if (S_ISREG (ent->mode)) {
		str[0] = '-';
	}
	else if (S_ISCHR (ent->mode)) {
		str[0] = 'c';
	}
	else if (S_ISBLK (ent->mode)) {
		str[0] = 'b';
	}
	else if (S_ISFIFO (ent->mode)) {
		str[0] = 'p';
	}
	else if (S_ISSOCK (ent->mode)) {
		str[0] = 's';
	}
	else {
		str[0] = '?';
	}

	str[1] = (ent->mode & S_IRUSR) ? 'r' : '-';
	str[2] = (ent->mode & S_IWUSR) ? 'w' : '-';
	if (ent->mode & S_ISUID) {
		str[3] = (ent->mode & S_IXUSR) ? 's' : 'S';
	}
	else {
		str[3] = (ent->mode & S_IXUSR) ? 'x' : '-';
	}

	str[4] = (ent->mode & S_IRGRP) ? 'r' : '-';
	str[5] = (ent->mode & S_IWGRP) ? 'w' : '-';
	if (ent->mode & S_ISGID) {
		str[6] = (ent->mode & S_IXGRP) ? 's' : 'S';
	}
	else {
		str[6] = (ent->mode & S_IXGRP) ? 'x' : '-';
	}

	str[7] = (ent->mode & S_IROTH) ? 'r' : '-';
	str[8] = (ent->mode & S_IWOTH) ? 'w' : '-';
	if (ent->mode & S_ISVTX) {
		str[9] = (ent->mode & S_IXOTH) ? 't' : 'T';
	}
	else {
		str[9] = (ent->mode & S_IXOTH) ? 'x' : '-';
	}

	str[10] = 0;
}

static
char slist_get_hex_digit (unsigned n)
{
	n &= 0x0f;

	if (n >= 10) {
		return ('a' + n - 10);
	}

	return ('0' + n);
}

static
int slist_quote_char (unsigned char c)
{
	if ((c <= 32) || (c >= 127)) {
		return (1);
	}

	if (c == '=') {
		return (1);
	}

	return (0);
}

static
char *slist_quote_name (const char *str)
{
	unsigned            i, j, n;
	const unsigned char *src;
	char                *dst;

	src = (const unsigned char *) str;

	i = 0;
	n = 0;
	while (src[i] != 0) {
		if (slist_quote_char (src[i])) {
			n += 3;
		}
		else {
			n += 1;
		}
		i += 1;
	}

	dst = malloc (n + 1);
	if (dst == NULL) {
		return (NULL);
	}

	i = 0;
	j = 0;
	while (src[i] != 0) {
		if (slist_quote_char (src[i])) {
			dst[j + 0] = '=';
			dst[j + 1] = slist_get_hex_digit (src[i] >> 4);
			dst[j + 2] = slist_get_hex_digit (src[i]);
			j += 3;
		}
		else {
			dst[j] = src[i];
			j += 1;
		}
		i += 1;
	}

	dst[j] = 0;

	return (dst);
}

static
const char *slist_get_display_name (slist_t *ent)
{
	if (par_noquote) {
		return (ent->name);
	}

	return (ent->qname);
}

static
const char *slist_get_display_link (slist_t *ent)
{
	if (par_noquote) {
		return (ent->link);
	}

	return (ent->qlink);
}

static
void slist_set_empty_md5 (char *str)
{
	strcpy (str, "--------------------------------");
}

static
int slist_get_md5 (slist_t *ent, char *str)
{
	FILE          *fp;
	md5_t         md5;
	unsigned      i;
	size_t        n;
	unsigned char buf[32];

	if (ent->link != NULL) {
		if ((par_link_md5 == 0) || !S_ISREG (ent->lmode)) {
			slist_set_empty_md5 (str);
			return (0);
		}
	}
	else if (ent->isdir || !S_ISREG (ent->mode)) {
		slist_set_empty_md5 (str);
		return (0);
	}

	fp = fopen (ent->name, "rb");
	if (fp == NULL) {
		return (1);
	}

	md5_init (&md5);

	n = fread (par_buf, 1, par_buf_size, fp);
	while (n > 0) {
		md5_calc (&md5, par_buf, n);

		n = fread (par_buf, 1, par_buf_size, fp);
	}

	if (ferror (fp)) {
		fclose (fp);
		return (1);
	}

	md5_done (&md5);

	fclose (fp);

	md5_get_sum (&md5, buf);

	for (i = 0; i < 16; i++) {
		*(str++) = slist_get_hex_digit (buf[i] >> 4);
		*(str++) = slist_get_hex_digit (buf[i] & 0x0f);
	}

	*str = 0;

	return (0);
}

static
void slist_print_short (slist_t *ent, FILE *fp)
{
	if (par_prefix) {
		fputs (par_prefix, fp);
	}

	if (ent->isdir) {
		if (strcmp (ent->name, "/") == 0) {
			fputs ("/\n", fp);
		}
		else {
			fprintf (fp, "%s/\n", slist_get_display_name (ent));
		}
	}
	else {
		fprintf (fp, "%s\n", slist_get_display_name (ent));
	}

	if (par_flush) {
		fflush (fp);
	}
}

static
void slist_update_stats (slist_t *ent)
{
	par_stat_entries += 1;

	if (ent->link != NULL) {
		par_stat_links += 1;
	}
	else if (ent->isdir) {
		par_stat_dirs += 1;
	}
	else if (S_ISREG (ent->mode)) {
		if (par_stat_files == 0) {
			par_stat_min = ent->size;
			par_stat_max = ent->size;
		}
		else {
			if (ent->size < par_stat_min) {
				par_stat_min = ent->size;
			}

			if (ent->size > par_stat_max) {
				par_stat_max = ent->size;
			}
		}

		par_stat_files += 1;
		par_stat_total += ent->size;
	}
}

static
int slist_print_entry (slist_t *ent, FILE *fp)
{
	char      mode[16];
	char      md5[64];
	long      tz;
	struct tm *tm;

	if (par_stats) {
		slist_update_stats (ent);
	}

	if (par_list_long == 0) {
		slist_print_short (ent, fp);
		return (0);
	}

	if (par_list_md5) {
		if (slist_get_md5 (ent, md5)) {
			fprintf (stderr, "%s: md5 failed (%s)\n",
				argv0, ent->qname
			);

			slist_set_empty_md5 (md5);
		}

		fprintf (fp, "%s ", md5);
	}

	slist_get_mode (ent, mode);
	fprintf (fp, "%s ", mode);

#ifdef HAVE_TIMEZONE
	if (par_utc) {
		tm = gmtime (&ent->mtime);
		tz = 0;
	}
	else {
		tm = localtime (&ent->mtime);
		tz = -timezone;
	}
#else
	tm = gmtime (&ent->mtime);
	tz = 0;
#endif

	if (tm->tm_isdst > 0) {
		tz += 3600;
	}

	fprintf (fp, "%04d-%02d-%02d %02d:%02d:%02d ",
		tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
		tm->tm_hour, tm->tm_min, tm->tm_sec
	);

	if (par_list_tz) {
		if (tz < 0) {
			fprintf (fp, "-%02ld:%02ld ",
				-tz / 3600, (-tz / 60) % 60
			);
		}
		else {
			fprintf (fp, "+%02ld:%02ld ",
				tz / 3600, (tz / 60) % 60
			);
		}
	}

	fprintf (fp, "%10llu ", ent->isdir ? 0LLU : ent->size);

	if (par_list_inode) {
		fprintf (fp, "%lld %lld ",
			(long long) ent->dev,
			(long long) ent->inode
		);
	}

	if (par_prefix != NULL) {
		fputs (par_prefix, fp);
	}

	if (ent->link != NULL) {
		fprintf (fp, "%s -> %s\n",
			slist_get_display_name (ent),
			slist_get_display_link (ent)
		);
	}
	else if (ent->isdir) {
		if (strcmp (ent->name, "/") == 0) {
			fputs ("/\n", fp);
		}
		else {
			fprintf (fp, "%s/\n", slist_get_display_name (ent));
		}
	}
	else {
		fprintf (fp, "%s\n", slist_get_display_name (ent));
	}

	if (par_flush || par_list_md5) {
		fflush (fp);
	}

	return (0);
}

static
int slist_printdir (slist_t *lst, unsigned long cnt, FILE *fp)
{
	int           r;
	unsigned long i;

	r = 0;

	for (i = 0; i < cnt; i++) {
		if (lst[i].isdir) {
			continue;
		}

		if (lst[i].link != NULL) {
			if (S_ISDIR (lst[i].lmode)) {
				if (par_list_dir_link == 0) {
					continue;
				}
			}
		}

		r |= slist_print_entry (&lst[i], fp);
	}

	fflush (fp);

	return (r);
}

static
void slist_print_stats (FILE *fp)
{
	unsigned long long avg;

	if (par_stat_files > 0) {
		avg = par_stat_total / par_stat_files;
	}
	else {
		avg = 0;
	}

	fprintf (fp, "# entries:      %lu\n", par_stat_entries);
	fprintf (fp, "# directories:  %lu\n", par_stat_dirs);
	fprintf (fp, "# symlinks:     %lu\n", par_stat_links);
	fprintf (fp, "# files:        %lu\n", par_stat_files);
	fprintf (fp, "# total size:   %llu\n", par_stat_total);
	fprintf (fp, "# average size: %llu\n", avg);
	fprintf (fp, "# minimum size: %llu\n", par_stat_min);
	fprintf (fp, "# maximum size: %llu\n", par_stat_max);
}

static
void slist_del_entry (slist_t *ent)
{
	free (ent->name);
	free (ent->link);
	free (ent->qname);
	free (ent->qlink);
}

static
void slist_del (slist_t **lst, unsigned long *cnt)
{
	unsigned long i;
	slist_t       *ent;

	ent = *lst;

	for (i = 0; i < *cnt; i++) {
		slist_del_entry (&ent[i]);
	}

	free (ent);

	*lst = NULL;
	*cnt = 0;
}

static
char *slist_fix_dir (const char *dir)
{
	unsigned i;
	char     *ret;

	while ((dir[0] == '.') && (dir[1] == '/')) {
		dir += 2;
	}

	if (*dir == 0) {
		return (strdup ("."));
	}

	ret = strdup (dir);
	if (ret == NULL) {
		return (NULL);
	}

	if (strcmp (ret, "/") == 0) {
		return (ret);
	}

	i = 0;
	while (ret[i] != 0) {
		i += 1;
	}

	if (i > 0) {
		while ((i > 0) && (ret[i - 1] == '/')) {
			i -= 1;
			ret[i] = 0;
		}

		if (i == 0) {
			ret[0] = '/';
			ret[1] = 0;
		}
	}

	return (ret);
}

static
char *slist_make_name (const char *dir, const char *name)
{
	char     *ret;
	unsigned i, n;

	if (*dir == 0) {
		return (strdup (name));
	}

	if (strcmp (dir, ".") == 0) {
		return (strdup (name));
	}

	n = strlen (dir);
	n += strlen (name);

	ret = malloc (n + 2);
	if (ret == NULL) {
		return (NULL);
	}

	i = 0;
	while (dir[i] != 0) {
		ret[i] = dir[i];
		i += 1;
	}

	while ((i > 0) && (ret[i - 1] == '/')) {
		i -= 1;
		ret[i] = 0;
	}

	ret[i++] = '/';

	while (*name != 0) {
		ret[i++] = *(name++);
	}

	ret[i] = 0;

	return (ret);
}

static
int slist_set_entry (slist_t *ent, const char *name, int delname)
{
	struct stat st;

	if (delname) {
		ent->name = (char *) name;
	}
	else {
		ent->name = strdup (name);
		if (ent->name == NULL) {
			return (1);
		}
	}

	if (lstat (ent->name, &st)) {
		free (ent->name);
		return (1);
	}

	ent->qname = slist_quote_name (ent->name);

	ent->isdir = S_ISDIR (st.st_mode);

	if (S_ISLNK (st.st_mode)) {
		ssize_t     n;
		struct stat stl;

		n = readlink (ent->name, par_buf, par_buf_size);
		if ((n < 0) || (n >= par_buf_size)) {
			free (ent->name);
			return (1);
		}

		ent->link = malloc (n + 1);
		if (ent->link == NULL) {
			free (ent->name);
			return (1);
		}

		memcpy (ent->link, par_buf, n);
		ent->link[n] = 0;

		if (stat (ent->name, &stl)) {
			ent->lmode = 0;
		}
		else {
			ent->lmode = stl.st_mode;

			if (par_follow_symlink) {
				if (S_ISDIR (stl.st_mode)) {
					ent->isdir = 1;
				}
			}
		}

		ent->qlink = slist_quote_name (ent->link);
	}
	else {
		ent->link = NULL;
		ent->qlink = NULL;
		ent->lmode = 0;
	}

	ent->mode = st.st_mode;
	ent->mtime = st.st_mtime;
	ent->size = st.st_size;
	ent->dev = st.st_dev;
	ent->inode = st.st_ino;

	return (0);
}

static
int slist_add_entry (slist_t **lst, unsigned long *cnt, const char *dir, const char *name)
{
	slist_t *ent;
	slist_t tmp;
	char    *str;

	if (strcmp (name, ".") == 0) {
		return (0);
	}

	if (strcmp (name, "..") == 0) {
		return (0);
	}

	str = slist_make_name (dir, name);
	if (str == NULL) {
		return (1);
	}

	if (slist_set_entry (&tmp, str, 1)) {
		return (1);
	}

	ent = realloc (*lst, (*cnt + 1) * sizeof (slist_t));
	if (ent == NULL) {
		slist_del_entry (&tmp);
		return (1);
	}

	ent[*cnt] = tmp;

	*lst = ent;
	*cnt += 1;

	return (0);
}

/*
 * Compare file names (case sensitive)
 */
static
int slist_cmp_name (slist_t *ent1, slist_t *ent2)
{
	return (strcmp (ent1->name, ent2->name));
}

/*
 * Compare file names (case insensitive)
 */
static
int slist_cmp_name_nocase (slist_t *ent1, slist_t *ent2)
{
	return (strcasecmp (ent1->name, ent2->name));
}

/*
 * Compare file name lengths
 */
static
int slist_cmp_name_length (slist_t *ent1, slist_t *ent2)
{
	unsigned n1, n2;

	n1 = strlen (ent1->name);
	n2 = strlen (ent2->name);

	if (n1 < n2) {
		return (-1);
	}
	else if (n1 > n2) {
		return (1);
	}

	return (0);
}

/*
 * Compare file sizes
 */
static
int slist_cmp_size (slist_t *ent1, slist_t *ent2)
{
	if (ent1->size < ent2->size) {
		return (-1);
	}
	else if (ent1->size > ent2->size) {
		return (1);
	}

	return (0);
}

/*
 * Compare modification times
 */
static
int slist_cmp_mtime (slist_t *ent1, slist_t *ent2)
{
	if (ent1->mtime < ent2->mtime) {
		return (-1);
	}
	else if (ent1->mtime > ent2->mtime) {
		return (1);
	}

	return (0);
}

/*
 * Compare inode numbers
 */
static
int slist_cmp_inode (slist_t *ent1, slist_t *ent2)
{
	if (ent1->inode < ent2->inode) {
		return (-1);
	}
	else if (ent1->inode > ent2->inode) {
		return (1);
	}

	return (0);
}

static
int slist_cmp_reverse (int v, int rev)
{
	if (rev) {
		v = -v;
	}

	if (par_reverse) {
		v = -v;
	}

	return (v);
}

/*
 * Compare two entries ent1 and ent2 according to the key str
 */
static
int slist_cmp_str (slist_t *ent1, slist_t *ent2, const char *str)
{
	int v;
	int rev;

	if (str == NULL) {
		v = slist_cmp_name (ent1, ent2);

		return (slist_cmp_reverse (v, 0));
	}

	rev = 0;

	while (*str != 0) {
		switch (*str) {
		case '+':
			rev = 0;
			break;

		case '-':
			rev = 1;
			break;

		case '#':
			v = slist_cmp_name_length (ent1, ent2);
			if (v) {
				return (slist_cmp_reverse (v, rev));
			}
			rev = 0;
			break;

		case 'i':
			v = slist_cmp_inode (ent1, ent2);
			if (v) {
				return (slist_cmp_reverse (v, rev));
			}
			rev = 0;
			break;

		case 'n':
			v = slist_cmp_name (ent1, ent2);
			if (v) {
				return (slist_cmp_reverse (v, rev));
			}
			rev = 0;
			break;

		case 'N':
			v = slist_cmp_name_nocase (ent1, ent2);
			if (v) {
				return (slist_cmp_reverse (v, rev));
			}
			rev = 0;
			break;

		case 's':
			v = slist_cmp_size (ent1, ent2);
			if (v) {
				return (slist_cmp_reverse (v, rev));
			}
			rev = 0;
			break;

		case 't':
			v = slist_cmp_mtime (ent1, ent2);
			if (v) {
				return (slist_cmp_reverse (v, rev));
			}
			rev = 0;
			break;

		}

		str += 1;
	}

	return (0);
}

/*
 * Compare two entries
 */
static
int slist_cmp (slist_t *ent1, slist_t *ent2)
{
	if ((ent1->isdir == 0) && (ent2->isdir == 0)) {
		return (slist_cmp_str (ent1, ent2, par_sort_files));
	}

	if (ent1->isdir && ent2->isdir) {
		return (slist_cmp_str (ent1, ent2, par_sort_dirs));
	}

	if (ent1->isdir) {
		return (-1);
	}

	return (1);
}

/*
 * Sort a directory
 */
static
void slist_sortdir (slist_t *lst, unsigned long cnt)
{
	unsigned long i, j;
	slist_t       tmp;

	for (i = 1; i < cnt; i++) {
		tmp = lst[i];
		j = i;
		while (j > 0) {
			if (slist_cmp (&tmp, &lst[j - 1]) >= 0) {
				break;
			}
			lst[j] = lst[j - 1];
			j -= 1;
		}

		if (j != i) {
			lst[j] = tmp;
		}
	}
}

static
int slist_readdir (slist_t **lst, unsigned long *cnt, const char *dir)
{
	DIR           *dp;
	char          *fix;
	struct dirent *de;

	fix = slist_fix_dir (dir);

	dp = opendir (fix);
	if (dp == NULL) {
		fprintf (stderr, "%s: can't open directory (%s)\n",
			argv0, dir
		);
		free (fix);
		return (1);
	}

	de = readdir (dp);
	while (de != NULL) {
		if (slist_add_entry (lst, cnt, fix, de->d_name)) {
			fprintf (stderr, "%s: error (%s / %s)\n",
				argv0, dir, de->d_name
			);
		}
		de = readdir (dp);
	}

	free (fix);

	closedir (dp);

	return (0);
}

static
int slist_dir (const char *dir)
{
	int           r;
	unsigned long i;
	unsigned long cnt;
	slist_t       *lst;

	r = 0;

	lst = NULL;
	cnt = 0;

	if (slist_readdir (&lst, &cnt, dir)) {
		return (1);
	}

	slist_sortdir (lst, cnt);

	if (par_files_first) {
		r |= slist_printdir (lst, cnt, stdout);
	}

	for (i = 0; i < cnt; i++) {
		if (lst[i].isdir) {
			if (par_list_dir) {
				slist_print_entry (&lst[i], stdout);
			}

			if (par_xdev) {
				if (lst[i].dev != par_xdev_dev) {
					continue;
				}
			}

			r |= slist_dir (lst[i].name);
		}
	}

	if (par_files_first == 0) {
		r |= slist_printdir (lst, cnt, stdout);
	}

	slist_del (&lst, &cnt);

	return (r);
}

static
int slist (const char *dir)
{
	slist_t ent;

	if (slist_set_entry (&ent, slist_fix_dir (dir), 1)) {
		return (1);
	}

	par_xdev_dev = ent.dev;

	if (ent.isdir == 0) {
		slist_print_entry (&ent, stdout);
		slist_del_entry (&ent);
		return (0);
	}

	if (par_list_dir) {
		if (strcmp (ent.name, ".") != 0) {
			slist_print_entry (&ent, stdout);
		}
	}

	slist_del_entry (&ent);

	return (slist_dir (dir));
}

int main (int argc, char *argv[])
{
	int i, r;
	int have_dirs;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("slist");
			return (EXIT_SUCCESS);
		}
	}

	tzset();

	have_dirs = 0;
	r = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "d", "directories")) {
			par_list_dir = 1;
		}
		else if (str_isarg2 (argv[i], "D", "no-dir-symlinks")) {
			par_list_dir_link = 0;
		}
		else if (str_isarg2 (argv[i], "f", "files-first")) {
			par_files_first = 1;
		}
		else if (str_isarg2 (argv[i], "F", "flush")) {
			par_flush = 1;
		}
		else if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing directory name\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			r |= slist (argv[i]);
			have_dirs = 1;
		}
		else if (str_isarg2 (argv[i], "I", "inodes")) {
			par_list_inode = 1;
		}
		else if (str_isarg2 (argv[i], "l", "long")) {
			par_list_long = 1;
		}
		else if (str_isarg2 (argv[i], "L", "symlink-md5")) {
			par_link_md5 = 1;
		}
		else if (str_isarg2 (argv[i], "m", "md5")) {
			par_list_md5 = 1;
		}
		else if (str_isarg2 (argv[i], "p", "prefix")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing prefix\n",
					argv0
				);
				return (EXIT_FAILURE);
			}
			free (par_prefix);
			par_prefix = strdup (argv[i]);
		}
		else if (str_isarg2 (argv[i], "q", "no-quote")) {
			par_noquote = 1;
		}
		else if (str_isarg2 (argv[i], "r", "reverse")) {
			par_reverse = 1;
		}
		else if (str_isarg2 (argv[i], "y", "follow-symlinks")) {
			par_follow_symlink = 1;
		}
		else if (str_isarg2 (argv[i], "s", "sort")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing sort string\n",
					argv0
				);
				return (EXIT_FAILURE);
			}
			par_sort_files = argv[i];
		}
		else if (str_isarg2 (argv[i], "S", "sort-directories")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing sort string\n",
					argv0
				);
				return (EXIT_FAILURE);
			}
			par_sort_dirs = argv[i];
		}
		else if (str_isarg2 (argv[i], "t", "stats")) {
			par_stats = 1;
		}
		else if (str_isarg2 (argv[i], "u", "utc")) {
			par_utc = 0;
		}
		else if (str_isarg2 (argv[i], "x", "same-device")) {
			par_xdev = 1;
		}
		else if (str_isarg2 (argv[i], "z", "timezone")) {
			par_list_tz = 1;
		}
		else if (argv[i][0] != '-') {
			r |= slist (argv[i]);
			have_dirs = 1;
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n",
				argv0, argv[i]
			);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (have_dirs == 0) {
		r |= slist (".");
	}

	if (par_stats) {
		fputs ("\n", stdout);
		slist_print_stats (stdout);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
