/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/tjoint/tjoint.c                                          *
 * Created:     1999-03-17 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 1999-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/file.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include "tjoint.h"


#define TJOINT_BUF_SIZE MISC_BUF_SIZE


static const char    *argv0 = NULL;

static unsigned      par_file_cnt = 0;
static file_t        *par_file = NULL;

static unsigned long par_buf_size = TJOINT_BUF_SIZE;

static int           par_append = 0;


static
void prt_help (void)
{
	fputs (
		"tjoint: split output to multiple files\n\n"
		"usage: tjoint [options] [[@]filename...]\n"
		"  -a, --append      Append to following output files [no]\n"
		"  -b, --buffer int  Set the buffer size\n",
		stdout
	);

	fflush (stdout);
}

static
file_t *file_new (void)
{
	par_file_cnt += 1;

	par_file = (file_t *) realloc (par_file, par_file_cnt * sizeof (file_t));
	if (par_file == NULL) {
		return (NULL);
	}

	return (par_file + par_file_cnt - 1);
}

static
int file_new_fd (int fd, int close)
{
	file_t *f;

	f = file_new();
	if (f == NULL) {
		return (1);
	}

	file_set_fd (f, fd, close);

	if (file_open_out (f)) {
		return (1);
	}

	return (0);
}

static
int file_new_fname (const char *fname)
{
	int    r;
	file_t *f;

	f = file_new();
	if (f == NULL) {
		return (1);
	}

	if (fname[0] == '@') {
		file_set_name (f, fname + 1);
		r = file_open_app (f);
	}
	else if (par_append) {
		file_set_name (f, fname);
		r = file_open_app (f);
	}
	else {
		file_set_name (f, fname);
		r = file_open_out (f);
	}

	return (r);
}

static
int do_tjoint (file_t *f_inp)
{
	unsigned      i;
	unsigned char buf[TJOINT_BUF_SIZE];
	unsigned long cnt;

	cnt = file_read (f_inp, buf, par_buf_size);

	while (cnt > 0) {
		for (i = 0; i < par_file_cnt; i++) {
			file_write_block (par_file + i, buf, cnt);
		}

		cnt = file_read (f_inp, buf, par_buf_size);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int    i, r;
	file_t f_inp;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help ();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("tjoint");
			return (EXIT_SUCCESS);
		}
	}

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "a", "append")) {
			par_append = 1;
		}
		else if (str_isarg2 (argv[i], "b", "buffer")) {
			i += 1;

			if (i >= argc) {
				fprintf (stderr, "%s: missing buffer size\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_buf_size)) {
				fprintf (stderr, "%s: illegal number (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}

			if (par_buf_size < 1) {
				fprintf (stderr,
					"%s: buffer size %lu too small, using 1 instead.\n",
					argv0, par_buf_size
				);
				par_buf_size = 1;
			}
			else if (par_buf_size > TJOINT_BUF_SIZE) {
				fprintf (stderr,
					"%s: buffer size %lu too large. using %lu instead.\n",
					argv0, par_buf_size,
					(unsigned long) TJOINT_BUF_SIZE
				);
				par_buf_size = TJOINT_BUF_SIZE;
			}
		}
		else if (strcmp (argv[i], "-") == 0) {
			file_new_fd (STDOUT_FILENO, 0);
		}
		else if (strcmp (argv[i], "--") == 0) {
			file_new_fd (STDERR_FILENO, 0);
		}
		else {
			r = file_new_fname (argv[i]);
			if (r) {
				fprintf (stderr, "%s: can't open file (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}

		i += 1;
	}

	file_set_stdin (&f_inp);

	if (file_open_inp (&f_inp)) {
		return (EXIT_FAILURE);
	}

	r = do_tjoint (&f_inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
